package Exec::Sudo;

use strict;
use warnings;

use Cwd qw(getcwd);

use fields qw(user);

use constant SUDO_BINARY => '/usr/bin/sudo';
use constant CHOWN_BINARY => '/bin/chown';
use constant ALLOWED_PLATFORMS => qw(linux);

if (not -x SUDO_BINARY) {
  die "sudo executable file ".SUDO_BINARY." is not found ";
}
if (not -x CHOWN_BINARY) {
  die "chown executable file ".CHOWN_BINARY." is not found ";
}
if (not grep($^O eq $_, ALLOWED_PLATFORMS)) {
  die "Current platform $^O cannot be used by ".__PACKAGE__." ";
}

sub new {
  my $self = shift;
  $self = fields::new($self) unless ref($self);
  return $self;
}

sub user {
  my $self = shift;
  if (scalar(@_) > 1) {
    require Carp;
    Carp::croak(__PACKAGE__."::user should be called with exactly one argument");
  } elsif (scalar(@_) == 1) {
    if ($_[0] =~ /^([-[:alnum:]_.]*)$/) {
      $self->{user} = $1;
    }
    else {
      require Carp;
      Carp::croak(__PACKAGE__."::user can accept username consisting of only alnum, '-', '_' and '.' characters");
    }
  }
  return $self->{user};  
}

sub system {
  my $self = shift;
  $self->user()
    or die "Cannot execute ".__PACKAGE__."::system when user is not defined ";

  local @ENV{qw(IFS CDPATH ENV BASH_ENV PATH)};
  delete @ENV{qw(IFS CDPATH ENV BASH_ENV PATH)};

  my @sudo_call;
  my $wd = quotemeta(getcwd());
  my $user;
  my $rc;
  $wd = quotemeta(getcwd());
  $wd =~ /^(.*)$/;
  $wd = $1;
  $user = $self->user();
  if ($< != $>) {
    die "UID is not equal to EUID,strange effects may occur, I won't work ";
  }
  @sudo_call = (SUDO_BINARY, '-u', $user, '-i', '--', '-c'); 

  system(SUDO_BINARY, '-i', '--', '-c', CHOWN_BINARY." -R $user $wd");
  $rc = $?;
  if ($rc) {
    warn "Cannot chown: $rc ";
    return $rc;
  }
  system(@sudo_call, join(' ', ("cd $wd;", map(quotemeta, @_))));
  $rc = $?;
  if ($rc) {
    warn "Cannot execute: $rc ";
    return $rc;
  }
  system(SUDO_BINARY, '-i', '--', '-c', CHOWN_BINARY." -R $< $wd");
  $rc = $?;
  if ($rc) {
    warn "Cannot chown: $rc ";
    return $rc;
  }
}  

1;
