package NodeDesc;

use strict;
use warnings;

use Sys::Hostname;
use XML::Writer;

use SystemHWDetection::Wsystem qw(get_cpu_desc getRAM getSystemArchitecture);

sub new
{
  my $class = shift;
  my $self  = {};

  $self->{CLIENT_ID} = hostname();
  $self->{CPU_TYPE} = getSystemArchitecture();
  $self->{RAM_SIZE} = getRAM();
  $self->{OS} = $^O;

  my @cpus = get_cpu_desc();
  $self->{NUM_CORES} = scalar(@cpus);
 
  my $cpu = shift @cpus;
  # $min_freq contains minimum processor frequency
  # $min_pref contains minimum processor performance (in flops)
  # $model and $vendor contains model and vendor name of CPU (core)
  #   with either minimum frequency or minimum performance.
  #   If minimums are reached on different CPU, $model and $vendor
  #   will contain model and vednor name of one of those which comes last.
  my $min_freq = $cpu->{FREQ};
  my $min_perf = $cpu->{PERF};
  my $model = $cpu->{MODEL};
  my $vendor = $cpu->{VENDOR};
  while ($cpu = shift @cpus) {
    if ($cpu->{FREQ} < $min_freq) {
      $min_freq = $cpu->{FREQ};
      $model = $cpu->{MODEL};
      $vendor = $cpu->{VENDOR};
    }
    if ($cpu->{PERF} < $min_perf) {
      $min_perf = $cpu->{PERF};
      $model = $cpu->{MODEL};
      $vendor = $cpu->{VENDOR};
    }
  }
  
  $self->{FREQ} = $min_freq;
  $self->{PERF} = $min_perf;
  $self->{MODEL} = $model;
  $self->{VENDOR} = $vendor;

  bless ($self, $class);
  return $self;
}

#
# It takes one argument of class XML::Writer and
# calls appropriate methods to write XML file describing client node.
#
sub write_xml
{
  my $self = shift;
  my $writer = shift;

  $writer->startTag('client', 'id' => $self->client_id);
  $writer->emptyTag('cpu',
    'type' => $self->{CPU_TYPE},
    'num-cores' => $self->{NUM_CORES},
    'freq' => $self->{FREQ},
    'perf' => $self->{PERF},
    'model' => $self->{MODEL},
    'vendor' => $self->{VENDOR}
  );
  $writer->emptyTag('ram', 'size' => $self->{RAM_SIZE});
  $writer->emptyTag('os', 'type' => $self->{OS});
  $writer->endTag('client');

  return $writer;
}

#
# Returns XML describing client node as string
#
sub xml_string
{
  my $self = shift;
  my $str;
  my $writer = XML::Writer->new(OUTPUT => \$str);
  $self->write_xml($writer);
  $writer->end();

  return $str;
}

sub client_id
{
  my $self = shift;
  if (@_) {$self->{CLIENT_ID} = shift;}
  return $self->{CLIENT_ID};
}

1;