package SystemHWDetection::Linux;

use strict;
use warnings;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(get_cpu_desc getRAM getSystemArchitecture);

our $VERSION = '0.1';

use SystemHWDetection::CpuMultiplier qw(getMultiplier);

use POSIX;

sub get_cpu_desc () {
 my $ok = 0;
 my $f = 0;
 my $h = 0;
 my @r=();
 my $line;
 my $elem = {};

 open(SESAME, "/proc/cpuinfo");
 while ($line = <SESAME>)
 {
  if ($ok == 3)
  {
    $ok = 0;
    $elem->{PERF} = $h * $f;
    push @r, $elem;
  }
   if ($line =~ /model\s+name\s+:\s*(\S.*\S)\s*$/){
     my $model = $1;
     $f=CpuMultiplier::getMultiplier($line);     
     $elem->{MODEL} = $model;
     $elem->{OPS_PER_CYCLE} = $f;
     $ok+=1;
   }
   elsif ($line =~ /vendor_id\s+:\s*(\S.*\S)\s*$/){
     my $vendor = $1;
     $elem->{VENDOR} = $vendor;
     $ok+=1;
   }
   elsif ($line =~ /cpu\s+MHz\s*:\s*([\d\.]+)/ix) {
    $h = $1 * 1000000;
    $elem->{FREQ} = $h;
    $ok+=1;
   }
   elsif ($line =~ /timer\s+frequency\s*:\s*([\d\.]+)/ix){
    $h = $1 * 1000000;
    $elem->{FREQ} = $h;
    $ok +=1;
   }
 }
 close (SESAME);
 return @r;
}
sub getRAM()
{
        my $mem;
	my $line;
	open(SESAME, "/proc/meminfo");
	while ($line = <SESAME>)
	{
		if ($line =~ /^memTotal:\s+(\d+)\s+kB\s*$/ix)
		{
			$mem = $1;
			last;
		}
	}
	close(SESAME);
	return $mem * 1024;
}
sub getSystemArchitecture()
{
	(my $a, my $b, my $c, my $d, my $m ) = POSIX::uname();
	if ($m =~ /^i\d86$/ix)
	{
		return "x86";
	}
	if ($m =~ /^mips$/ix)
	{
		return "mips";
	}
	if ($m =~ /^alpha$/ix)
	{
		return "alpha";
	}
	if ($m =~ /^powerpc$/ix)
	{
		return "powerpc";
	}
	if ($m =~ /^ia64$/ix)
	{
		return "ia64";
	}
	if ($m =~ /^x86_64$/ix)
	{
		return "x86_64";
	}
	return "unknown";	
}
1;
