package SystemHWDetection::Win32;

use strict;
use warnings;

require Exporter;

our @ISA = qw(Exporter);
our @EXPORT_OK = qw(get_cpu_desc getRAM getSystemArchitecture);

our $VERSION = 0.1;

use SystemHWDetection::CpuMultiplier qw(getMultiplier);

use Win32::OLE;
use Win32::Registry; 
use File::Glob;


sub get_cpu_desc () {
 my @r=();
 my $ok = 1;
 my $f = 0;
 my $h = 0;
 my $line;
 my $tips;
 my $num = 0;
 my ($type, $value);
 my $elem = {};

 while ($ok) {
  $h = 0;
  $ok = 0;
  if ($::HKEY_LOCAL_MACHINE->Open("HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\$num", $tips)) {$ok=1;}
  if ($ok) {                                                                                         
   $tips->QueryValueEx("~MHz", $type, $value);# or die "No ~MHz: $^E";
   if ($value) { $h = $value; }
   $h = (0 + $h) * 1000000; 
   $elem->{FREQ} = $h;
   $tips->QueryValueEx("VendorIdentifier", $type, $line);
   $line =~ s/^\s*(\S.*\S)\s*$/$1/;
   $elem->{VENDOR} = $line;
   $tips->QueryValueEx("ProcessorNameString", $type, $line);
   $line =~ s/^\s*(\S.*\S)\s*$/$1/;
   $elem->{MODEL} = $line;
   $f=CpuMultiplier::getMultiplier($line);     
   $elem->{OPS_PER_CYCLE} = $f;
   $elem->{PERF} = $f * $h;
   $num+=1;
   push @r, $elem;
   
  } 
 }
 return @r;
} 
sub getRAM ()
{
  my $instances = 
  Win32::OLE->GetObject("winmgmts:{impersonationLevel=impersonate,(Debug)}")->InstancesOf("win32_PhysicalMemory")  
    or die "Cannot access WMI";
  my $mem=0;
  for my $inst (in $instances) { 
    $mem+=$inst->{Capacity};
  } 
  return $mem;
}     
1;

sub getSystemArchitecture()
{
  my $instances =
  Win32::OLE->GetObject("winmgmts:{impersonationLevel=impersonate,(Debug)}")->InstancesOf("win32_Processor")  
    or die "Cannot access WMI";
  for my $inst (in $instances) {
    my $arc = $inst->{Architecture};
    if ($arc==0)
    {
        return "x86";
    }
    if ($arc==1)
    {
        return "MIPS";
    }
    if ($arc==2)
    {
        return "Alpha";
    }
    if ($arc==3)
    {
        return "PowerPC";
    }
    if ($arc==6)
    {
        return "IA-64";
    }
    if ($arc==9)
    {
        return "x86-64";
    }
    return "unknown";
  }
}
