package Task::Portion;

use strict;
use warnings;

require File::Copy;
use File::Spec;
use File::Temp;
use IO::File;
use MIME::Parser;

use Util qw(untaint_path);

use URLList;
our @ISA = qw(URLList);

sub new
{
  my $class = shift;
  my $task = shift;
  my $elem = shift;
  
  my $self = $class->parse_xml_object($elem);
  $self->{TASK} = $task;
  $self->{FILENAME} = undef;
  
  bless ($self, $class);
  return $self;
}

#
# Creates Portion from XML::Parser object pointing to portion-server element
#
sub parse_xml_object
{
  my $class = shift;
  my $elem = shift;

  ref($elem) =~ /::portion-server$/ or die "Not portion-server XML element";

  my $self = $class->SUPER::parse_xml_object($elem);

  bless ($self, $class);
  return $self;
}

sub filename
{
  my $self = shift;
  if(@_) {
    my $filename = shift;
    $self->{FILENAME} = untaint_path($filename);
  }
  return $self->{FILENAME}; 
}

sub task
{
  my $self = shift;
  return $self->{TASK};
}

# XXX cur_* methods should be moved to separate class representing single portion, not potion server
sub cur_num
{
  my $self = shift;
  return $self->{CUR_NUM};
}

sub cur_taskarg
{
  my $self = shift;
  return $self->{CUR_TASKARG};
}

#
# Unlinks portion file using name from the description in the specified
# diretory
#
sub unlink
{
  my $self = shift;

  my $fname = $self->filename();
  unlink($fname);
}

#
# Fetches portion data file from URL
#
sub fetch
{
  my $self = shift;

  $self->unlink;
  my $file = $self->filename;

  my $req = Task::Portion::Request->new(
    $self,
    $self->get_next_url()
  );

  my $tmp = File::Temp->new(DIR => $self->task->client->configurator->workdir());

  my $ua = $self->task->ua();
  $ua->add_handler(response_header =>
    sub {
      my ($response, $ua, $h) = @_;
      $response->{default_add_content} = 0;
      print $tmp $response->headers->as_string;
      print $tmp "\n";
      binmode $tmp;
    }
  );
  $ua->add_handler(response_data =>
    sub {
      my ($response, $ua, $h, $data) = @_;
      print $tmp $_[3];
    }
  );
  $ua->add_handler(response_done =>
    sub {
      $tmp->close;
    }
  );

  my $resp = $ua->request($req);

  $ua->remove_handler("response_header");
  $ua->remove_handler("response_data");
  $ua->remove_handler("response_done");
  
  $resp->is_success or die "Cannot fetch portion from " . $req->uri() . ": " . $resp->status_line;

  my $mparser = MIME::Parser->new;
  $mparser->output_under($self->task->client->configurator->workdir());
  my $ent = $mparser->parse_open($tmp->filename());

  # XXX current portion info should be moved to separate class
  my $parser = XML::Parser->new(Style => 'Objects', Pkg => 'Task::XML');
  my $portion_desc_parsed = $parser->parse($ent->parts(0)->bodyhandle->as_string);
  $self->{CUR_NUM} = $portion_desc_parsed->[0]->{Kids}->[0]->{num};
  $self->{CUR_TASKARG} = $portion_desc_parsed->[0]->{Kids}->[0]->{args};

  File::Copy::move($ent->parts(1)->bodyhandle->path, $file)
    or die "Cannot move data portion to path $file: $!";

  $mparser->filer->purge;
}

package Task::Portion::Request;

use strict;
use warnings;

use HTTP::Request;
use XML::Writer;

our @ISA = qw(HTTP::Request);

#
# Creates new request
# Arguments are:
# - Reference to Portion
# - URL to make request
#
sub new
{
  my $class = shift;
  my $portion = shift;
  my $url = shift;

  my $self = $class->SUPER::new(POST => $url);

  my $task_id = $portion->task()->task_id();
  my $task_name = $portion->task()->task_name();
  my $session_id = $portion->task()->session_id();
  my $client_id = $portion->task()->client()->configurator()->nodedesc->client_id;

  my $req_body;
  my $writer = XML::Writer->new(OUTPUT => \$req_body);
  $writer->emptyTag('get-portion',
    'client-id' => $client_id,
    'task-name' => $task_name,
    'task-id' => $task_id,
    'session-id' => $session_id
  );
  $writer->end;

  $self->content_type('text/xml');
  $self->content($req_body);

  return $self;
}

1;
