package Task::Result;

use strict;
use warnings;

use File::Spec;
use IO::File;

use Util qw(untaint_path);

use URLList;
our @ISA = qw(URLList);

sub new
{
  my $class = shift;
  my $task = shift;
  my $elem = shift;
  
  my $self = $class->parse_xml_object($elem);
  $self->{TASK} = $task;
  $self->{FILENAME} = undef;
  
  bless ($self, $class);
  return $self;
}

#
# Creates Result from XML::Parser object pointing to result-server element
#
sub parse_xml_object
{
  my $class = shift;
  my $elem = shift;

  ref($elem) =~ /::result-server$/ or die "Not result-server XML element";

  my $self = $class->SUPER::parse_xml_object($elem);

  bless ($self, $class);
  return $self;
}

sub filename
{
  my $self = shift;
  if(@_) {
    my $filename = shift;
    $self->{FILENAME} = untaint_path($filename);
  }
  return $self->{FILENAME}; 
}

sub task
{
  my $self = shift;
  return $self->{TASK};
}

#
# Unlinks result file using name from the description in the specified
# diretory
#
sub unlink
{
  my $self = shift;

  my $fname = $self->filename();
  unlink($fname);
}

#
# Posts result data file from specified directory (using name
#   from file description)
# Arguments are:
# - Directory name to take file from. (XXX - unused???)
#
sub post_from_dir
{
  my $self = shift;
  my $dir = shift or die "Directory name is not defined";

  my $req = Task::Result::Request->new(
    $self,
    $self->get_next_url(),
    $self->filename
  );
  my $resp = $self->task()->ua()->request($req);
  $resp->is_success or die "Cannot post result to " . $req->uri();
}

package Task::Result::Request;

use strict;
use warnings;

use File::stat;
use HTTP::Request;
use HTTP::Request::Common;
use XML::Writer;

our @ISA = qw(HTTP::Request);

#
# Creates new request
# Arguments are:
# - Reference to Result
# - URL to make request
# - file handle to take result from
#
sub new
{
  my $class = shift;
  my $result = shift;
  my $url = shift;
  my $filename = shift;

  my $self;
#  my $self = $class->SUPER::new(POST => $url);
  
  my $task_name = $result->task()->task_name();
  my $task_id = $result->task()->task_id();
  my $session_id = $result->task()->session_id();
  my $client_id =
    $result->task()->client()->configurator()->nodedesc->client_id;
  my $portion_num = $result->task()->portion()->cur_num();
  my $time = $result->task->clipart()->work_time();

  my $req_body;
  my $writer = XML::Writer->new(OUTPUT => \$req_body);
  $writer->startTag('post-result',
    'client-id' => $client_id,
    'task-name', $task_name,
    'task-id' => $task_id,
    'session-id' => $session_id,
# XXX- number of portions is always 1
    'q' => 1
  );
  $writer->emptyTag('result-data',
    'size' => stat($filename)->size(),
    'num' => $portion_num,
    'time' => $time
  );
  $writer->endTag('post-result');
  $writer->end;

  local $HTTP::Request::Common::DYNAMIC_FILE_UPLOAD = 1;
  $self = POST $url,
    Content_Type => 'form-data',
    Content => [ '1' => $req_body,
                 '2' => [$filename]
               ];

  bless ($self, $class);
  return $self;
}

1;
