package URLList;

use strict;
use warnings;

#
# First arg is 'fallback' or random'
# Other arguments are URLs
#
sub new
{
  my $class = shift;
  my $self  = {};

  my $order = shift;
  $order = lc $order;
  if(defined $order) {
    if($order eq 'fallback' or $order eq 'random') {
      $self->{ORDER} = $order;
    } else {
      die "Order argument must one of 'fallback' or 'random'";
    }
  }

  if(@_) {
    $self->{LIST} = [@_];
  } else {
    $self->{LIST}  = [];
  }
   
  bless ($self, $class);
  return $self;
}

#
# Creates URLList from XML::Parser object style object
#
sub parse_xml_object
{
  my $class = shift;
  my $elem = shift;

  my $order;
  my @urls;

  $order = $elem->{order};
  for my $child (@{$elem->{Kids}}) {
    # Skip empty text block
    next if (ref($child) =~ /::Characters$/ and $child->{Text} =~ /^\s*$/s);

    ref($child) =~ /::url$/ or die "Not an url child";
    push @urls, $child->{Kids}->[0]->{Text};

  }

  my $self = __PACKAGE__->new($order, @urls);
  bless($self, $class);
  return $self;
}

sub order
{
  my $self = shift;
  if(@_) {
    if($_[0] eq 'fallback' or $_[0] eq 'random') {
      $self->{ORDER} = shift;
    } else {
      die "Order argument must one of 'fallback' or 'random'";
    }
  }
  return $self->{ORDER};
}

sub urls
{
  my $self = shift;
  if(@_) {$self->{LIST} = [@_];}
  return @{$self->{LIST}}; 
}

sub clear_urls
{
  my $self = shift;
  $self->{LIST} = [];
}

sub add_url
{
  my $self = shift;
  push @{$self->{LIST}}, @_;
}

sub get_next_url
{
  my $self = shift;
  return ${$self->{LIST}}[0];
}

1;
