#!/usr/bin/perl -T

use strict;
use warnings;

use FindBin;

use Cwd qw(getcwd);

my $startdir;

BEGIN
{
  $FindBin::Bin =~ /^(.*)$/;
  require lib;
  lib->import($1);

  getcwd =~ /^(.*)$/;
  $startdir = $1;
}

use Configurator;
use Client;

use Carp;

# It is needed to make END block work in case of signal caught
use sigtrap 'handler' => \&signal_handler, 'normal-signals';

our $debug_on_exit = sub {};

sub signal_handler {
  my $sig = shift;
  $debug_on_exit->($sig);
  exit(254);
}

main();

sub main
{
  my $configurator;
  $configurator = Configurator->new();
  if($configurator->debug()) {
    $SIG{__DIE__} = \&confess;
    $debug_on_exit = \&Carp::cluck;
    $Carp::Verbose = 1;
    my $client = Client->new(CONFIGURATOR => $configurator);
    $client->run();
  } elsif ($configurator->max_portions()) {
    my $client = Client->new(CONFIGURATOR => $configurator);
    $client->run();
  } else {
# XXX - quick and dirty hack for client to restart itself  when not in debug mode.
#       Makes it very hard to stop it with SIGTERM
    while (1) {
      eval {
        my $client = Client->new(CONFIGURATOR => $configurator);
        $client->run();
      };
      if ($@ =~ /^Caught a SIGINT at /) {
        die $@;
      } elsif ($@ =~ /^Caught a SIGHUP at /) {
        die $@;
      } elsif ($@ =~ /^Caught a SIGTERM at /) {
        die $@;
      } elsif ($@) {
        warn "We died: $@. Restarting";
	chdir $startdir;
        require File::Temp;
        File::Temp::cleanup();
      }
      sleep 10;
    }
  }
}
