#!/usr/bin/perl

#
# Common Perl modules
#

use strict;
use warnings;

#use Carp;
#$Carp::Verbose = 1;
#$SIG{__WARN__} = \&confess;
#$SIG{__DIE__} = \&confess;

use FindBin qw($Bin);
#use lib "$Bin/../pm";
#use lib "$Bin/../tasks";
use sigtrap qw(die normal-signals error-signals);

use Time::HiRes;
use POSIX ":sys_wait_h";
use File::Basename;
#use File::Spec;
#use File::Spec::Functions;
#use Data::Dumper;

#
# Our own modules
#

use XServAPI;
use XServUtils;
use XServProcessor;
use XServSubserver;

#
# Constants
#

*PROXY_IN_BUFFER_MAX = \20;
*PROXY_IN_BUFFER_MIN = \10;
*PROXY_OUT_BUFFER_MAX = \20;
*PROXY_OUT_BUFFER_MIN = \10;

my %tAPI;
$tAPI{Perl} = $tAPI{Files} = $tAPI{Subserver} = $tAPI{Proxy} = 1;
my ($HTTP_PREFIX, $PROXY_HTTP_PREFIX) = ('http://', 'http://');
my $pid_ts;
my $PID = $$;

#
# Common variables
#

our ($ts_prHost, $ts_prPort, $ts_ssHost, $ts_ssPort, $ts_taskId, $ts_taskName, $ts_taskAPI, $ts_taskList, $ts_sessId, $ts_portServer, $ts_resServer, $ts_taskPath);
our ($ts_prSock, $ts_ssASWHost, $ts_ssASWPort, $ts_ssREQHost, $ts_ssREQPort);
our ($ts_proxyASWServer, $ts_proxyREQServer, $ts_ssTempFiles, $ts_ssReqOnDisk, $ts_ssTempDir);
our ($ts_taskArgs, $ts_taskIn, $ts_taskOut, $ts_taskDesc, $ts_dlOrder, $ts_dlSites_s, $ts_logFile, $ts_outExtension, $ts_logOldFormat, $ts_logPath);
our (@ts_dlSites, %ts_dlFiles);
our ($ts_upHost, $ts_upPort, $ts_upServer, $ts_proxyTmp, $ts_inBufferMax, $ts_inBufferMin, $ts_outBufferMax, $ts_outBufferMin, $ts_clId, $ts_clOS, $ts_clCPUType, $ts_clCPUCores, $ts_clCPUFreq, $ts_clCPUPerf, $ts_clRAMSize, $ts_httpProxy, $ts_proxyType);
our ($StartTime, $StartTimeMsec) = Time::HiRes::gettimeofday();
our ($ts_inBufferSize, $ts_outBufferSize) = (0, 0);
our ($ts_keepFirst, $ts_singlePortion, $ts_portionWait, $ts_xCliAPI, $ts_xUser, $ts_xCliInitCmd, $ts_xCliPortionCmd, $ts_xCliPortionIn, $ts_xCliPortionOut, $ts_debugShowXML, $ts_xFile);
our ($ts_useSSL, $ts_SSL_key_file, $ts_SSL_cert_file, $ts_SSL_trusted_dir);
our ($proxy_taskArchive);
our ($ts_queueRefresh, $ts_queueRetries, $ts_queueNP, $ts_queueTimeLimit);
our ($ts_cleoQueue, $ts_cleoGetInfo, $ts_cleoXMLStatus, $ts_cleoRawStatus, $ts_cleoProfile, $ts_cleoPreCmd, $ts_cleoTimeLimit);
our ($ts_llClass, $ts_llNodes, $ts_llTPN);
our ($ts_uniTargetSystem, $ts_uniNodes, $ts_uniCPUs);

our @TaskFilesNames;
our @TaskFilesArgs;
our (@InBufferPortions, @OutBufferPortions);

#
# Zombies processing
#

#sub REAPER
#{
#  1 until (-1 == waitpid(-1, WNOHANG));
#  $SIG{CHLD} = \&REAPER;
#}
#$SIG{CHLD} = \&REAPER;
$SIG{CHLD} = 'IGNORE';

sub GetIniSettings
{
  my $ini = $_[0];
  my %IniSettings=();
  open (FP, "< $ini") or die ("XServ: can't read $ini!\n");
  while (<FP>)
  {
    chomp;
    s/#.*//;
    s/^\s+//;
    s/\s+$//;
    next unless length;
    my ($var, $value) = split(/\s*=\s*/, $_, 2);
    $IniSettings{$var} = $value;
  }
  close (FP);
  if (!defined($IniSettings{proxyType}) || $IniSettings{proxyType} eq 'X-Com')
  {
    if (!defined($IniSettings{prSock}))
    {
      $ts_prHost=$IniSettings{prHost} or die ("XServ: (prHost) processor host name is missing in $ini!\n");
      $ts_prPort=$IniSettings{prPort} or die ("XServ: (prPort) processor port number is missing in $ini!\n");
    }
    else 
    {
      $ts_prSock = $IniSettings{prSock};
    }
    $ts_ssHost=$IniSettings{ssHost} or die ("XServ: (ssHost) subserver host name is missing in $ini!\n");
    $ts_ssPort=$IniSettings{ssPort} or die ("XServ: (ssPort) subserver port number is missing in $ini!\n");
  }
  $ts_ssASWHost = (defined $IniSettings{ssASWHost} ? $IniSettings{ssASWHost} : $ts_ssHost);
  $ts_ssASWPort = (defined $IniSettings{ssASWPort} ? $IniSettings{ssASWPort} : $ts_ssPort);
  $ts_ssREQHost = (defined $IniSettings{ssREQHost} ? $IniSettings{ssREQHost} : $ts_ssHost);
  $ts_ssREQPort = (defined $IniSettings{ssREQPort} ? $IniSettings{ssREQPort} : $ts_ssPort);

  $ts_taskAPI=$IniSettings{taskAPI} or die ("XServ: (taskAPI) task API is missing in $ini!\n"); 
  die ("XServ: (taskAPI) incorrect task API in $ini!\n") if !exists($tAPI{$ts_taskAPI});
  $ts_keepFirst = (defined $IniSettings{keepFirst} ? $IniSettings{keepFirst} : ($ts_taskAPI eq 'Proxy' ? 0 : 1)); # or die ("XServ: (keepFirst) portions strategy is missing in $ini!\n");
  $ts_singlePortion = (defined $IniSettings{singlePortion} ? $IniSettings{singlePortion} : 0);
  $ts_portionWait = (defined $IniSettings{portionWait} ? $IniSettings{portionWait} : 0);
  $ts_xCliAPI = (defined $IniSettings{xCliAPI} ? $IniSettings{xCliAPI} : 'Perl');
  $ts_debugShowXML = (defined $IniSettings{debugShowXML} ? $IniSettings{debugShowXML} : 0);
  $ts_ssTempFiles = (defined $IniSettings{ssTempFiles} ? $IniSettings{ssTempFiles} : 0);
  $ts_ssTempFiles = 0 if ( (defined ($ts_prSock)) && (($ts_taskAPI eq 'Files') || ($ts_taskAPI eq 'Proxy')) );
  $ts_ssTempDir = (defined $IniSettings{ssTempDir} ? $IniSettings{ssTempDir} : 'tmp');
  $ts_ssReqOnDisk = (defined $IniSettings{ssReqOnDisk} ? $IniSettings{ssReqOnDisk} : 0);
  $ts_useSSL = (defined $IniSettings{useSSL} ?  $IniSettings{useSSL} : undef);
#  if (defined ($ts_useSSL) && ($ts_useSSL =~/^noverify$/ix || $ts_useSSL =~/^verifyClientCertificate$/ix))
#  {
#    $ts_SSL_key_file = $IniSettings{SSLkey} or die ("XServ: (SSLkey) SSL configuration is missing in $ini!\n");
#    $ts_SSL_cert_file = $IniSettings{SSLcert} or die ("XServ: (SSLcert) SSL configuration is missing in $ini!\n");
#    $ts_SSL_trusted_dir = $IniSettings{SSLdir} or die ("XServ: (SSLdir) SSL configuration is missing in $ini!\n");
#    $HTTP_PREFIX = "https://";
#  }
  if (defined ($ts_useSSL) && ($ts_useSSL =~/^noverify$/ix || $ts_useSSL =~/^verify$/ix || $ts_useSSL =~/^verifycrl$/ix))
  {
    $ts_SSL_key_file = $IniSettings{SSLkey} or die ("XServ: (SSLkey) SSL configuration is missing in $ini!\n");
    $ts_SSL_cert_file = $IniSettings{SSLcert} or die ("XServ: (SSLcert) SSL configuration is missing in $ini!\n");
    if ($ts_useSSL=~/^noverify$/ix)
    {
      $ts_SSL_trusted_dir = '';
    } else {
      $ts_SSL_trusted_dir = $IniSettings{SSLdir} or die ("XServ: (SSLdir) SSL configuration is missing in $ini!\n");
    }
    $HTTP_PREFIX = "https://";
  }
  if ($ts_xCliAPI eq 'Processes')
  {
    # $ts_xUser = (defined $IniSettings{xUser} ? $IniSettings{xUser} : (defined($ENV{USER}) ? $ENV{USER} : (defined ($ENV{USERNAME}) ? $ENV{USERNAME} : '')));
    $ts_xUser = (defined $IniSettings{xUser} ? $IniSettings{xUser} : undef );
    $ts_xCliInitCmd = (defined $IniSettings{xCliInitCmd} ? $IniSettings{xCliInitCmd} : '');
    $ts_xCliPortionCmd = (defined $IniSettings{xCliPortionCmd} ? $IniSettings{xCliPortionCmd} : '');
    $ts_xCliPortionIn = (defined $IniSettings{xCliPortionIn} ? $IniSettings{xCliPortionIn} : 'portion.in');
    $ts_xCliPortionOut = (defined $IniSettings{xCliPortionOut} ? $IniSettings{xCliPortionOut} : 'portion.out');
  }
  if (($ts_taskAPI ne 'Proxy') && ($ts_taskAPI ne 'Subserver'))
  {
    $ts_taskName=$IniSettings{taskName} or die ("XServ: (taskName) task name is missing in $ini!\n");
    $ts_taskArgs = (defined $IniSettings{taskArgs} ? $IniSettings{taskArgs} : '');
    if (defined $IniSettings{xFile}) 
    { 
      # $ts_taskArgs = $IniSettings{xFile}.' '.$ts_taskArgs;
      $ts_xFile = $IniSettings{xFile};
    }
    $ts_taskDesc = (defined $IniSettings{taskDesc} ? $IniSettings{taskDesc} : '');
    $ts_taskId = (defined $IniSettings{taskId} ? $IniSettings{taskId} : 1);
    $ts_taskPath = (defined $IniSettings{taskPath} ? $IniSettings{taskPath} : $Bin.'/../tasks/'.$ts_taskName);
  }
  @ts_dlSites = (defined $IniSettings{dlSites} ? split (/ /, $IniSettings{dlSites}) : (defined($ts_ssHost) && defined($ts_ssPort) ? $HTTP_PREFIX."$ts_ssHost:$ts_ssPort/" : ''));
  $ts_dlSites_s = @ts_dlSites > 1 ? join (' ', @ts_dlSites) : $ts_dlSites[0];
  $ts_dlOrder = (defined $IniSettings{dlOrder} ? $IniSettings{dlOrder} : 'Fallback');
  if (defined $IniSettings{dlFiles})
  {
    foreach my $t (split(/ /, $IniSettings{dlFiles}))
    {
      $ts_dlFiles{basename($t)}=dirname($t);
    }
  }
  if ($ts_taskAPI eq 'Files')
  {
    $ts_taskIn = (defined $IniSettings{taskIn} ? $IniSettings{taskIn} : '');
    $ts_taskOut=$IniSettings{taskOut} or die ("XServ: (taskOut) output path is missing in $ini!\n");
    $ts_taskList=$IniSettings{taskList} or die ("XServ: (taskList) input file list is missing in $ini!\n");
    $ts_outExtension = (defined $IniSettings{outExtension} ? $IniSettings{outExtension} : '.out');
  }
  $ts_logPath = (defined $IniSettings{logPath} ? $IniSettings{logPath} : 'log');
  if (!defined $IniSettings{logFile})
  {
    my @tTime = localtime($StartTime);
    $ts_logFile=$ts_logPath.'/'.($tTime[5]+1900).'-'.($tTime[4] < 9 ? '0' : '').($tTime[4]+1).'-'.($tTime[3] < 10 ? '0' : '').$tTime[3].'_'.($tTime[2] < 10 ? '0' : '').$tTime[2].'-'.($tTime[1] < 10 ? '0' : '').$tTime[1].'-'.($tTime[0] < 10 ? '0' : '').$tTime[0].'.'.$StartTimeMsec;
  }
  else { $ts_logFile = $IniSettings{logFile} }
  $ts_logOldFormat = defined ($IniSettings{logOldFormat});
  if ($ts_taskAPI eq 'Proxy')
  {
    # $ts_upHost=$IniSettings{upHost} or die ("XServ: (upHost) upper server host name is missing in $ini!\n");
    # $ts_upPort=$IniSettings{upPort} or die ("XServ: (upPort) upper server port number is missing in $ini!\n");
    if (defined ($IniSettings{upServer}))
    {
      $ts_upServer = $IniSettings{upServer};
    }
    elsif (defined ($IniSettings{upHost}) && defined ($IniSettings{upPort}))
    {
      $ts_upHost = $IniSettings{upHost};
      $ts_upPort = $IniSettings{upPort};
      $ts_upServer = $PROXY_HTTP_PREFIX.$ts_upHost.':'.$ts_upPort;
    }
    else
    { die ("XServ: upper sever settings are missed in $ini!\n") }
    $ts_proxyTmp = (defined $IniSettings{proxyTmp} ? $IniSettings{proxyTmp} : '../proxy');
    $ts_inBufferMax = (defined $IniSettings{inBufferMax} ? $IniSettings{inBufferMax} : $::PROXY_IN_BUFFER_MAX);
    $ts_inBufferMin = (defined $IniSettings{inBufferMin} ? $IniSettings{inBufferMin} : $::PROXY_IN_BUFFER_MIN);
    $ts_outBufferMax = (defined $IniSettings{outBufferMax} ? $IniSettings{outBufferMax} : $::PROXY_OUT_BUFFER_MAX);
    $ts_outBufferMin = (defined $IniSettings{outBufferMin} ? $IniSettings{outBufferMin} : $::PROXY_OUT_BUFFER_MIN);
    $ts_clId=$IniSettings{clId} or die ("XServ: (clId) proxy-client Id is missing in $ini!\n");
    $ts_clOS=$IniSettings{clOS} or die ("XServ: (clOS) proxy-client OS type is missing in $ini!\n"); # linux MSWin32
    $ts_clCPUType=$IniSettings{clCPUType} or die ("XServ: (clCPUType) proxy-client CPU type is missing in $ini!\n"); # x86 x86_64 ia64
    $ts_clCPUCores=$IniSettings{clCPUCores} or die ("XServ: (clCPUCores) proxy-client CPU cores number is missing in $ini!\n");
    $ts_clCPUFreq=$IniSettings{clCPUFreq} or die ("XServ: (clCPUFreq) proxy-client CPU frequency is missing in $ini!\n"); # hertz
    $ts_clCPUPerf=$IniSettings{clCPUPerf} or die ("XServ: (clCPUPerf) proxy-client CPU performance is missing in $ini!\n"); # flops
    $ts_clRAMSize=$IniSettings{clRAMSize} or die ("XServ: (clRAMSize) proxy-client RAM size is missing in $ini!\n"); # bytes
    $ts_httpProxy=$IniSettings{httpProxy} if defined ($IniSettings{httpProxy});
    $ts_proxyType = (defined $IniSettings{proxyType} ? $IniSettings{proxyType} : 'X-Com');
    if ( ($ts_proxyType eq 'Cleo') || ($ts_proxyType eq 'Torque') || ($ts_proxyType eq 'LoadLeveler') || ($ts_proxyType eq 'Unicore'))
    {
      $ts_queueRefresh = (defined $IniSettings{queueRefresh} ? $IniSettings{queueRefresh} : 1);
      $ts_queueRetries = (defined $IniSettings{queueRetries} ? $IniSettings{queueRetries} : 5);
      $ts_queueNP = (defined $IniSettings{queueNP} ? $IniSettings{queueNP} : 1);
      $ts_queueTimeLimit = (defined $IniSettings{queueTimeLimit} ? $IniSettings{queueTimeLimit} : undef);
    }
    if ($ts_proxyType eq 'Cleo')
    {
      $ts_cleoQueue = (defined $IniSettings{cleoQueue} ? $IniSettings{cleoQueue} : 'regular');
      $ts_cleoXMLStatus = (defined $IniSettings{cleoXMLStatus} ? $IniSettings{cleoXMLStatus} : "/tmp/cleo-xml-status.$ts_cleoQueue");
      $ts_cleoRawStatus = (defined $IniSettings{cleoRawStatus} ? $IniSettings{cleoRawStatus} : "/tmp/cleo-status.$ts_cleoQueue"); 
      $ts_cleoProfile = (defined $IniSettings{cleoProfile} ? $IniSettings{cleoProfile} : '');
      $ts_cleoPreCmd = (defined $IniSettings{cleoPreCmd} ? $IniSettings{cleoPreCmd} : '');
      $ts_cleoGetInfo = (defined $IniSettings{cleoGetInfo} ? $IniSettings{cleoGetInfo} : 'Raw');
    }
    elsif ($ts_proxyType eq 'Torque')
    {
    }
    elsif ($ts_proxyType eq 'LoadLeveler')
    {
      $ts_llClass = (defined $IniSettings{llClass} ? $IniSettings{llClass} : undef);
      $ts_llNodes = (defined $IniSettings{llNodes} ? $IniSettings{llNodes} : undef);
      $ts_llTPN = (defined $IniSettings{llPPN} ? $IniSettings{llPPN} : $ts_queueNP);
    }
    elsif ($ts_proxyType eq 'Unicore')
    {
      $ts_uniTargetSystem = (defined $IniSettings{uniTargetSystem} ? $IniSettings{uniTargetSystem} : undef);
      $ts_uniNodes = (defined $IniSettings{uniNodes} ? $IniSettings{uniNodes} : undef);
      $ts_uniCPUs = (defined $IniSettings{uniCPUs} ? $IniSettings{uniCPUs} : $ts_queueNP);
    }
  }
}

#
# Main code starts here
#

die ("XServ: ini file is missing!\n") if (@ARGV != 1);
GetIniSettings ($ARGV[0]);

print STDERR "XServ: starting as ".($ts_taskAPI eq 'Perl' || $ts_taskAPI eq 'File' ? "common server ($ts_taskAPI)" : $ts_taskAPI.($ts_taskAPI eq 'Proxy' ? " ($ts_proxyType mode)" : ''))."\n";
print STDERR " Processor:\t\t".(defined $ts_prSock ? $ts_prSock : "$ts_prHost:$ts_prPort")."\n Subservers:\t\t$ts_ssHost:$ts_ssPort\n Download prefix:\t{@ts_dlSites}\n" if (!defined($ts_proxyType) || ($ts_proxyType eq 'X-Com'));
print STDERR " SSL mode:\t\t$ts_useSSL\n SSLkey:\t\t$ts_SSL_key_file\n SSLcert:\t\t$ts_SSL_cert_file\n SSLdir:\t\t$ts_SSL_trusted_dir\n" if defined $ts_useSSL;
if (($ts_taskAPI ne 'Proxy') && ($ts_taskAPI ne 'Subserver'))
{ print STDERR " Task API:\t\t$ts_taskAPI\n Task name:\t\t$ts_taskName\n Task ID:\t\t$ts_taskId\n Task arguments:\t$ts_taskArgs\n\n" }
elsif ($ts_taskAPI eq 'Proxy')
{ print STDERR " Upper server:\t\t$ts_upServer\n Input buffer:\t\t$ts_inBufferMax/$ts_inBufferMin\n Output buffer:\t\t$ts_outBufferMax/$ts_outBufferMin\n Proxy-client:\t\t$ts_clId ($ts_clOS/$ts_clCPUType/$ts_clCPUCores/$ts_clCPUFreq/$ts_clRAMSize)\n\n" }
 elsif ($ts_taskAPI eq 'Subserver')
{  print STDERR " Task API:\t\t$ts_taskAPI\n\n"; }
print STDERR "XServ: ready at $HTTP_PREFIX$ts_ssHost:$ts_ssPort\n" if (!defined($ts_proxyType) || $ts_proxyType eq 'X-Com');

unlink $ts_logFile if -f $ts_logFile;
open (LOG, ">> $ts_logFile") or die ("XServ: can't write to $ts_logFile!");

if ($ts_taskAPI ne 'Proxy')
{
  if (!$ts_logOldFormat)
  {
    XServUtils::LogWrite2($ARGV[0],'START','Server='.(defined($ts_prHost)?$ts_prHost:$ts_ssHost).'&STime='.Time::HiRes::time().  '&TaskAPI='.$ts_taskAPI.($ts_taskAPI eq 'Subserver' ? '' : '&TaskName='.$ts_taskName.'&TaskId='.$ts_taskId.'&TaskArgs={'  .$ts_taskArgs.'}'));
  }
  else
  {
    XServUtils::LogWrite($ARGV[0],"server_name=".$ts_prHost);
    XServUtils::LogWrite($ARGV[0],"task=".$ts_taskName);
    XServUtils::LogWrite($ARGV[0],"task_arg=".$ts_taskArgs);
  }
}

XServAPI::Initialize();

if (($ts_taskAPI eq 'Perl') || ($ts_taskAPI eq 'Files'))
{
  $pid_ts = fork();
  die ("XServ: can't fork: $!\n") if !defined($pid_ts);
  if ($pid_ts==0) # child process - Processor
  {
    my $ret = XServProcessor::main();
    XServAPI::Finalize();
    if ($ret != 1) { print STDERR "Terminated by timer.\n" }

  }
  else # parent processes - Subserver
  {
    XServSubserver::main();
  }
  onExit();
  kill 9, $pid_ts;
  POSIX::_exit(0) if ($^O eq 'MSWin32');  # rude exit on Windows
  exit(0);

}
elsif ($ts_taskAPI eq 'Proxy')
{
  if ($ts_proxyType eq 'X-Com')
  {
    $pid_ts = fork();
    die ("XServ: can't fork: $!\n") if !defined($pid_ts);
    if ($pid_ts!=0) # parent processes - Processor
    {
      while (1)
      {
        XServProcessor::main();
	XServAPI::Initialize();
      }
    }
    else # child process - Subserver
    {
      XServSubserver::main();
    }
  }
  elsif ( ($ts_proxyType eq 'Cleo') || ($ts_proxyType eq 'Torque') || ($ts_proxyType eq 'LoadLeveler') || ($ts_proxyType eq 'Unicore') )
  {
    require XServProxyQueue;
    XServProxyQueue::main();
  }
}
elsif ($ts_taskAPI eq 'Subserver')
{
  XServSubserver::main();
}

sub onExit
{
  print STDERR "XServ: cleaning... ";
  if (defined ($ts_xFile)) { unlink ($ts_xFile) }
  if (defined ($ts_prSock) && (-S $ts_prSock)) { unlink ($ts_prSock) }
  if (defined ($ts_ssTempDir) && (-d $ts_ssTempDir)) { unlink <$ts_ssTempDir/*> }
  if (defined ($ts_proxyTmp) && (-d $ts_proxyTmp))
  {
    my $proxyDir = "$ts_proxyTmp/$ts_taskName.$ts_taskId.$ts_sessId";
    if (-d "$proxyDir/in")
    { unlink <$proxyDir/in/*>; rmdir "$proxyDir/in" }
    if (-d "$proxyDir/out")
    { unlink <$proxyDir/out/*>; rmdir "$proxyDir/out" }
    if (-d "$proxyDir/work")
    { unlink <$proxyDir/work/*>; rmdir "$proxyDir/work" }
    unlink <$proxyDir/*>;
    rmdir $proxyDir;
  }
  print STDERR "OK\n";
}

END { onExit() if $PID == $$ }
