package XServAPI;

use File::Copy;
use File::Basename;
no strict 'refs';

# use Data::Dumper;

sub Initialize
{
  if ($::ts_taskAPI eq 'Perl')
  {
    # eval 'use '.$::ts_taskName.'::'.$::ts_taskName;
    my $pm = $::ts_taskPath.'/'.$::ts_taskName.'.pm';
    die ("XServAPI(Initialize): can't find task server part $pm!\n") unless -f $pm;
    require $pm;
    $::ts_taskName->import();
    # ($::ts_taskName.'::initialize')->($::ts_taskArgs);
    initialize ($::ts_taskArgs);
  }
  elsif ($::ts_taskAPI eq 'Files')
  {
    open (FileIn, "< $::ts_taskList")
      or die ("XServAPI(Initialize): can't open $::ts_taskList!\n");
    while (defined (my $line = <FileIn>))
    {
      chomp $line;
      my @tmp = split (/ /, $line, 2);
      $::TaskFilesNames[$#main::TaskFilesNames+1]=$tmp[0];
      $::TaskFilesArgs[$#main::TaskFilesArgs+1]=$tmp[1];
    }
    close (FileIn);
  }
  elsif ($::ts_taskAPI eq 'Proxy')
  {
    require XServProxy;
    XServProxy::proxyInit();
  }
}

sub GetFirstPortionNumber
{
  # if ($::ts_taskAPI eq 'Perl') { return ($::ts_taskName."::getFirstPortionNumber")->() }
  if ($::ts_taskAPI eq 'Perl') { return getFirstPortionNumber() }
  elsif ($::ts_taskAPI eq 'Files') { return 1 }
  elsif ($::ts_taskAPI eq 'Proxy') { return 1 }
}

sub GetLastPortionNumber
{
  # if ($::ts_taskAPI eq 'Perl') { return ($::ts_taskName."::getLastPortionNumber")->() }
  if ($::ts_taskAPI eq 'Perl') { return getLastPortionNumber() }
  elsif ($::ts_taskAPI eq 'Files') { return @main::TaskFilesNames }
  elsif ($::ts_taskAPI eq 'Proxy') { return 0 }
}

sub GetPortionArgs
{
  my $prtn = $_[0];
  if ($::ts_taskAPI eq 'Perl') { return $::ts_taskArgs }
  elsif ($::ts_taskAPI eq 'Files' ) { return $::TaskFilesNames[$prtn-1].' '.$::ts_taskArgs.(defined $::TaskFilesArgs[$prtn-1] ? ' '.$::TaskFilesArgs[$prtn-1] : '') }
  elsif ($::ts_taskAPI eq 'Proxy' )
  {
    my $fn = "$::ts_proxyTmp/$::ts_taskName.$::ts_taskId.$::ts_sessId/in/$prtn.arg";
    open F, "< $fn" 
      or die("XServAPI(GetPortionArgs): can't read from $fn!\n");
    my $args = <F>;
    close F;
    return $args;
  }
}

sub GetPortion
{
  my $prtn = $_[0];
  my $fh;
  # if ($::ts_taskAPI eq 'Perl') { open ($fh, "< ", \($::ts_taskName."::getPortion")->($prtn)) }
  if ($::ts_taskAPI eq 'Perl') { open ($fh, "< ", \getPortion($prtn)) }
  elsif ($::ts_taskAPI eq 'Files') 
  {
    my $fn = ($::ts_taskIn eq '' ? $::TaskFilesNames[$prtn-1] : $::ts_taskIn.'/'.$::TaskFilesNames[$prtn-1]);
    open ($fh, "< $fn")
      or die ("XServAPI(GetPortion): can't open $fn!\n");
  }
  elsif ($::ts_taskAPI eq 'Proxy') 
  {
    my $fn = "$::ts_proxyTmp/$::ts_taskName.$::ts_taskId.$::ts_sessId/in/$prtn.dat";
    open ($fh, "< $fn")
      or die ("XServAPI(GetPortion): can't read from $fn!\n");
    $::ts_inBufferSize--;
  }
  binmode $fh;
  return $fh;
}

sub GetPortionInFile
{
  my $prtn = $_[0];
  if ($::ts_taskAPI eq 'Perl')
  {
    # Not implemented yet
    die ("XServAPI(GetPortionInFile): Perl API is not implemented yet!\n");
  }
  elsif ($::ts_taskAPI eq 'Files')
  {
    my $fn = ($::ts_taskIn eq '' ? $::TaskFilesNames[$prtn-1] : $::ts_taskIn.'/'.$::TaskFilesNames[$prtn-1]);
    return $fn;
  }
  elsif ($::ts_taskAPI eq 'Proxy')
  {
    $::ts_inBufferSize--;
    return "$::ts_proxyTmp/$::ts_taskName.$::ts_taskId.$::ts_sessId/in/$prtn.dat";
  }
}

sub AddPortion
{
  my ($prtn, $data) = @_;
  # if ($::ts_taskAPI eq 'Perl') { return ($::ts_taskName."::addPortion")->($prtn, $$data) }
  if ($::ts_taskAPI eq 'Perl') { return addPortion ($prtn, $$data) }
  else
  {
    my $fout;
    if ($::ts_taskAPI eq 'Files') { $fout = $::ts_taskOut."/".basename($::TaskFilesNames[$prtn-1]).$::ts_outExtension }
    elsif ($::ts_taskAPI eq 'Proxy')
    { 
      $fout = "$::ts_proxyTmp/$::ts_taskName.$::ts_taskId.$::ts_sessId/out/$prtn.dat";
      $::ts_outBufferSize++;
      push (@::OutBufferPortions, $prtn);
    } 
    open (FP, "> $fout") or die ("XServAPI(AddPortion): can't write to $fout!\n");
    binmode FP;
    print FP $$data;
    close FP;
  }
}

sub AddPortionFromFile
{
  my ($prtn, $file) = @_;
  if ($::ts_taskAPI eq 'Perl')
  {
    # Not implemented yet
    die ("XServAPI(AddPortionFromFile): Perl API is not implemented yet!\n");
  }
  else
  {
    my $fout;
    if ($::ts_taskAPI eq 'Files') { $fout = $::ts_taskOut."/".basename($::TaskFilesNames[$prtn-1]).$::ts_outExtension }
    elsif ($::ts_taskAPI eq 'Proxy')
    {
      $fout = "$::ts_proxyTmp/$::ts_taskName.$::ts_taskId.$::ts_sessId/out/$prtn.dat";
      $::ts_outBufferSize++;
      push (@::OutBufferPortions, $prtn);
    }
    move ($file, $fout) or die ("XServAPI(AddPortionFromFile): can't move $file to $fout!\n");
  }
}

sub IsFinished
{
  # if ($::ts_taskAPI eq 'Perl') { return ($::ts_taskName."::isFinished")->() }
  if ($::ts_taskAPI eq 'Perl') { return isFinished () }
  elsif ($::ts_taskAPI eq 'Files') { return 0 }
  elsif ($::ts_taskAPI eq 'Proxy') { return 0 }
}

sub Finalize
{
  if ($::ts_taskAPI eq 'Perl')
  {
    finalize ();
  }
}

1;