package XServProxyQueue;

use strict;
use warnings;

our $uTaskPath;
our %QueuedPortions = ();
my %QueuedPortionsTime = ();
my %PortionsCounter = ();
my %QueuedPortionsCounter = ();
my $inQueue = 0;

sub main
{
  if ($::ts_proxyType eq 'Cleo')
  { require XServProxyQueueCleo; XServProxyQueueCleo->import(); print "XServProxyQueue: Cleo module loaded\n" }
  elsif ($::ts_proxyType eq 'Torque')
  { require XServProxyQueueTorque; XServProxyQueueTorque->import(); print "XServProxyQueue: Torque module loaded\n" }
  elsif ($::ts_proxyType eq 'LoadLeveler')
  { require XServProxyQueueLL; XServProxyQueueLL->import(); print "XServProxyQueue: LoadLeveler module loaded\n" }
  elsif ($::ts_proxyType eq 'Unicore')
  { require XServProxyQueueUnicore; XServProxyQueueUnicore->import(); print "XServProxyQueue: Unicore module loaded\n" }
  elsif ($::ts_proxyType eq 'Slurm')
  { require XServProxyQueueSlurm; XServProxyQueueSlurm->import(); print "XServProxyQueue: SLURM module loaded\n" }

  my $queue = queue_name ();
  # $uTaskPath = $::ts_proxyTmp.'/'.$::ts_taskName.'.'.$::ts_taskId;
  $uTaskPath = $XServProxy::proxyDir;
  chdir $uTaskPath;
  system ("tar xfz $uTaskPath/$::proxy_taskArchive");
  if (queue_init() != 1)
  {
    print STDERR "XServProxyQueue: unable to initialize queue!\n";
    return 0;
  }

  my ($cStart, $cEnd);
  while (1)
  {
#print STDERR "QueuedPortions: {"."@{[ %QueuedPortions ]}"."}\n";
    $cStart = time();
    my $fStart = Time::HiRes::time();
    print "XServProxyQueue: checking queue '$queue'... ";
    my %FoundTasks = queue_found_tasks ();
    my @tmp1 = keys (%FoundTasks);
    my $fEnd = Time::HiRes::time();
    print @tmp1." our tasks\n";
    XServUtils::LogWrite2 ($::ts_clId, 'CHECK', "STime=$fStart&ETime=$fEnd&Queued=".@tmp1);
#print STDERR "FoundTasks: {"."@{[ %FoundTasks ]}"."}\n";
    foreach my $tid (keys %QueuedPortions)
    {
      unless (exists $FoundTasks{$tid})
      {
	my $prt = $QueuedPortions{$tid};
        push (@::OutBufferPortions, $prt);
	push (@::OutBufferPortionsPath, "$XServProxyQueue::uTaskPath/out/$prt.$QueuedPortionsCounter{$tid}");
	delete ($QueuedPortions{$tid});
	my $ct = Time::HiRes::time();
	$QueuedPortionsTime{$prt} = $ct - $QueuedPortionsTime{$prt};
	helper_rm ($prt);
	portion_time_write ($prt, $QueuedPortionsCounter{$tid});
	$inQueue--;
	my $tt = int($QueuedPortionsTime{$prt});
	my $wt = my $wti = 0;
	if ( -f "$uTaskPath/out/$prt.$QueuedPortionsCounter{$tid}.wt" )
	{
	  open WH, "$uTaskPath/out/$prt.$QueuedPortionsCounter{$tid}.wt";
	  $wt = <WH>;
	  $wti = int ($wt);
	  close WH;
	}
	print "XServProxyQueue: portion $prt left queue '$queue' in $tt ($wti+".($tt-$wti).") s.\n";
	my $fsize = ( -f "$uTaskPath/out/$prt.$QueuedPortionsCounter{$tid}.dat" ? -s "$uTaskPath/out/$prt.$QueuedPortionsCounter{$tid}.dat" : 0 );
	XServUtils::LogWrite2 ($::ts_clId, 'READY', "Portion=$prt&TaskId=$tid&ETime=$ct&TotalTime=$QueuedPortionsTime{$prt}&WorkingTime=$wt&WaitingTime=".($QueuedPortionsTime{$prt}-$wt)."&Size=$fsize");
      }
    }
#print STDERR "OutBufferPortions = ".@::OutBufferPortions.", ts_outBufferMin = $::ts_outBufferMin\n";
    while (@::OutBufferPortions >= $::ts_outBufferMax)
    {
      my $tStart = Time::HiRes::time();
      print "XServProxyQueue: sending $::ts_outBufferMin of ".@::OutBufferPortions." ready portion".(@::OutBufferPortions != 1 ? 's' : '')."\n";
      XServProxy::proxySendPortions ($::ts_outBufferMin, "$uTaskPath/out");
      my $tEnd = Time::HiRes::time();
      XServUtils::LogWrite2 ($::ts_clId, 'ASW', "STime=$tStart&ETime=$tEnd&Sent=$::ts_outBufferMin");
    }
#print STDERR "InBufferPortions = ".@::InBufferPortions.", ts_inBufferMax = $::ts_inBufferMax\n";
    while (@::InBufferPortions < $::ts_inBufferMin)
    {
      my $tStart = Time::HiRes::time();
      my $ptr = $::ts_inBufferMax-@::InBufferPortions;
      print "XServProxyQueue: ".@::InBufferPortions." portion".(@::InBufferPortions != 1 ? 's' : '')." left in input buffer, getting $ptr portion".($ptr != 1 ? 's' : '')."\n";
      XServProxy::proxyGetPortions ($ptr, "$uTaskPath/in");
      my $tEnd = Time::HiRes::time();
      XServUtils::LogWrite2 ($::ts_clId, 'REQ', "STime=$tStart&ETime=$tEnd&Received=$ptr");
    }
#print STDERR "inQueue = $inQueue, ts_outBufferMax = $::ts_outBufferMax\n";
    # while ($inQueue < $::ts_inBufferMax)
    while ($inQueue < $::ts_queueTasksIn)
    {
      my $p = shift (@::InBufferPortions);
      if (defined($p))
      {
	$PortionsCounter{$p} = ( exists ($PortionsCounter{$p}) ? $PortionsCounter{$p} + 1 : 1 );
	my $tStart = Time::HiRes::time();
        my $tid = queue_helper ($p, $PortionsCounter{$p});
	my $tEnd = Time::HiRes::time();
        # if ($tid > 0)
	if ( ($tid =~ /^https/) || ($tid > 0) )
        {
          $QueuedPortions{$tid} = $p;
	  $QueuedPortionsTime{$p} = Time::HiRes::time();
	  $QueuedPortionsCounter{$tid} = $PortionsCounter{$p};
          $inQueue++;
	  print "XServProxyQueue: portion $p queued in '$queue' with id $tid".($PortionsCounter{$p} > 1 ? " ($PortionsCounter{$p})" : '')."\n";
	  XServUtils::LogWrite2 ($::ts_clId, 'SUBMIT', "STime=$tStart&ETime=$tEnd&Portion=$p&TaskId=$tid");
        }
	else { print STDERR "XServProxyQueue: portion $p was not queued!\n" }
      }
      else { print STDERR "XServProxyQueue: oops, input buffer is empty!\n"; last }
    }
    $cEnd = time();
    if ($cEnd - $cStart < $::ts_queueRefresh)
    { sleep ($::ts_queueRefresh - ($cEnd - $cStart)) }
  }
}

sub helper_rm
{
  my ($p) = @_;
  my $helper = helper_name ($p);
  unlink <$helper*>;
}

sub helper_name
{
  my ($p) = @_;
  return $::ts_taskName.'.'.$p;
}

sub portion_time_write
{
  my ($p, $pc) = @_;
  open T, "> $uTaskPath/out/$p.$pc.tt";
  print T $QueuedPortionsTime{$p};
  close T;
}

sub portion_getargs
{
  my ($p) = @_;
  open ARG, "< $uTaskPath/in/$p.arg";
  my $args = <ARG>;
  close ARG;
  return $args;
}

1;
