package XServProxyQueueCleo;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(queue_init queue_found_tasks queue_helper queue_name);

use strict;
use warnings;

sub queue_init
{ 
  mkdir "$XServProxyQueue::uTaskPath/work" if $::ts_xCliAPI eq 'Processes';
  return 1;
}

sub queue_found_tasks
{
  my %ft = ();
  if ($::ts_cleoGetInfo eq 'XML')
  {
    open CLEO, $::ts_cleoXMLStatus;
    while (my $line = <CLEO>)
    {
      foreach (keys %XServProxyQueue::QueuedPortions)
      {
        $ft{$_} = 1 if $line =~ /task id=\"$_\"/;
      }
    }
    close CLEO;
  }
  elsif ($::ts_cleoGetInfo eq 'Raw')
  {
    open CLEO, $::ts_cleoRawStatus;
    while (my $line = <CLEO>)
    {
      foreach (keys %XServProxyQueue::QueuedPortions)
      {
        $ft{$_} = 1 if $line =~ /^\[task\.$_\]$/;
      }
    }
    close CLEO;
  }
  else { die "XServGateCleo: unknown method $::ts_cleoGetInfo for cleoGetInfo!\n" }
  return %ft;
}

sub queue_helper
{
  my ($p, $pc) = @_;
  my $helper = XServProxyQueue::helper_name ($p);
  my $args = XServProxyQueue::portion_getargs ($p);
  my $tl = ( (defined $::ts_queueTimeLimit) && ($::ts_queueTimeLimit =~ /(\d{1,2}):(\d{1,2}):(\d{1,2})/) ? $1*60*60+$2*60+$3 : undef );
  my $ch = "#!/bin/sh\n";
  if ($::ts_xCliAPI eq 'Perl')
  {
    $ch .= "perl -mTime::HiRes -e \"my \\\$start=Time::HiRes::time(); eval {do 'gctask'; gctask ('$::ts_taskName','$args','$p','$XServProxyQueue::uTaskPath/in/$p.dat','$XServProxyQueue::uTaskPath/out/$p.$pc.dat')}; open WH, '> $XServProxyQueue::uTaskPath/out/$p.$pc.wt'; print WH Time::HiRes::time()-\\\$start; close WH\"\n";
  }
  elsif ($::ts_xCliAPI eq 'Processes')
  {
    $ch .= "mkdir $XServProxyQueue::uTaskPath/work/$p
cp -l * $XServProxyQueue::uTaskPath/work/$p
cd $XServProxyQueue::uTaskPath/work/$p
cp -l $XServProxyQueue::uTaskPath/in/$p.dat $::ts_xCliPortionIn
./$::ts_xCliPortionCmd
cp $::ts_xCliPortionOut $XServProxyQueue::uTaskPath/out/$p.$pc.dat
cd $XServProxyQueue::uTaskPath
rm $XServProxyQueue::uTaskPath/work/$p/*
rmdir $XServProxyQueue::uTaskPath/work/$p
";
  }
  open CH, "> $helper";
  print CH $ch;
  close CH;
  chmod 0700, $helper;
  my $cmdstr = ($::ts_cleoPreCmd ne '' ? $::ts_cleoPreCmd."; " : '')."mpirun -stdin /dev/null -stdout /dev/null -stderr /dev/null -q $::ts_cleoQueue -np $::ts_queueNP ".($::ts_cleoProfile ne '' ? "-as $::ts_cleoProfile " : '').(defined $tl ? "-l $tl " : '')."./$helper";
#print STDERR $cmdstr."\n";
  my $cr = $::ts_queueRetries;
  my $id = 0;
  do
  {
    my $run = `$cmdstr`;
    if ($run =~ /Successfully added to queue $::ts_cleoQueue \(ID=(\d+)\)/)
    { $id = $1 }
    else { print STDERR "helper: mpirun said: $run".($cr > 1 ? "helper: retrying...\n" : '') }
    $cr--;
  } while ( ($id == 0) && ($cr > 0) );
  return $id;
}

sub queue_name
{
  return $::ts_cleoQueue;
}

1;