package XServProxyQueueUnicore;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT = qw(queue_init queue_found_tasks queue_helper queue_name);

use strict;
use warnings;
use File::Copy;

sub queue_init
{
  print "queue_init: checking availability of $::ts_uniTargetSystem\n";
  my $ucc = `ucc list-sites`;
  return 0 unless $ucc =~ /$::ts_uniTargetSystem/;
  my $homedir = "u6://$::ts_uniTargetSystem/Home/$::ts_taskName.$::ts_taskId";
  print "queue_init: preparing directory $homedir\n";
  system ("ucc mkdir $homedir");
  print "queue_init: uploading $homedir/$::proxy_taskArchive\n";
  system ("ucc put-file -s $::proxy_taskArchive -t $homedir/$::proxy_taskArchive");
  print "queue_init: unpacking files on site\n";
  open USH, "> untgz_helper.sh";
  print USH "#!/bin/bash
cd ".'$'."HOME/$::ts_taskName.$::ts_taskId
tar xfz $::proxy_taskArchive
";
  close USH;
  open UU, "> untgz_helper.u";
  print UU "{
#  Application: \"untgz_helper\",
  Executable: \"/bin/bash\",
  Imports: [
    { From: \"untgz_helper.sh\", To: \"untgz_helper.sh\" }
  ],
  Arguments: [\"untgz_helper.sh\"],
  Name: \"untgz_helper\",
  Description: \"Unpacking script for $::proxy_taskArchive\",
}
";
  close UU;
  system ("ucc run -O /dev/null -E /dev/null -s $::ts_uniTargetSystem untgz_helper.u");
  print "queue_init: done\n";
  return 1;
}

sub queue_found_tasks
{
  my %ft = ();
  foreach (keys %XServProxyQueue::QueuedPortions)
  {
    my $ucc = `ucc get-status $_`;
    if ($ucc =~ /^SUCCESSFUL/)
    {
      system ("ucc get-output $_");
      $_ =~ /res\=(.+)$/;
      my $uid = $1;
      my $p = $XServProxyQueue::QueuedPortions{$_};
      copy ("$uid.stdout", "out/$p.dat");
    }
    else
    {
      $ft{$_} = 1;
    }
  }
  return %ft;
}

sub queue_helper
{

  my ($p, $pc) = @_;
  my $helper_u = XServProxyQueue::helper_name($p).".u";
  my $helper_sh = XServProxyQueue::helper_name($p).".sh";
  my $args = XServProxyQueue::portion_getargs ($p);
  my $runtime;
  if ( (defined ($::ts_queueTimeLimit)) && ($::ts_queueTimeLimit =~ /(\d{1,2}):(\d{1,2}):(\d{1,2})/) )
  { $runtime = $1*60*60 + $2*60 + $3 }
  open HU, "> $helper_u";
  print HU "{
#  ApplicationName: \"$::ts_taskName\",
  Executable: \"/bin/bash\",
  Imports: [
    { From: \"$helper_sh\", To: \"$helper_sh\" },
    { From: \"in/$p.dat\", To: \"$p.in\" }
  ],
#  Exports: [
#   { From: \"$p.out\", To: \"out/$p.dat\" }
#  ],
  Arguments: [\"$helper_sh\"],
  Resources: { ".(defined $::ts_uniNodes ? "\n    Nodes: $::ts_uniNodes," : '').(defined $::ts_uniCPUs ? "\n    CPUs: $::ts_uniCPUs," : '').(defined $runtime ? "\n    Runtime: $runtime," : '')."
  },
  Name: \"$::ts_taskName.$::ts_taskId.$p\",
  Description: \"X-Com task $::ts_taskName.$::ts_taskId, portion $p\",
}
";
  close HU;

  my $ch = "#!/bin/sh
cp -l ".'$'."HOME/$::ts_taskName.$::ts_taskId/* .
";
  if ($::ts_xCliAPI eq 'Perl')
  {
    # $ch .= "perl -mTime::HiRes -e \"my \\\$start=Time::HiRes::time(); eval {do 'gctask'; gctask ('$::ts_taskName','$args','$p','$XServProxyQueue::uTaskPath/in/$p.dat','$XServProxyQueue::uTaskPath/out/$p.$pc.dat')}; open WH, '> $XServProxyQueue::uTaskPath/out/$p.$pc.wt'; print WH Time::HiRes::time()-\\\$start; close WH\"\n";
    $ch .= "perl -e \"eval {do 'gctask'; gctask ('$::ts_taskName','$args','$p','$p.in','$p.out')}\"
cat $p.out
";
  }
  elsif ($::ts_xCliAPI eq 'Processes')
  {
    $ch .= "cp -l $p.in $::ts_xCliPortionIn
./$::ts_xCliPortionCmd
cp -l $::ts_xCliPortionOut $p.out
cat $p.out
";
  }
  open HSH, "> $helper_sh";
  print HSH $ch;
  close HSH;
  chmod 0700, $helper_sh;

  my $cmdstr = "ucc run -b -a -s $::ts_uniTargetSystem $helper_u";
#print STDERR $cmdstr."\n";
  my $cr = $::ts_queueRetries;
  my $id = '';
  do
  {
    my $run = `$cmdstr`;
    if (-f 'job')
    { 
      open J, 'job';
      while (my $s = <J>)
      {
        if ($s =~ /^  "epr": "(.+)"/)
        { $id = $1; last }
      }
      close J;
    }
    else { print STDERR "queue_helper: ucc said: $run".($cr > 1 ? "queue_helper: retrying...\n" : '') }
    $cr--;
  } while ( ($id eq '') && ($cr > 0) );
  return $id;
}

sub queue_name
{
  return 'default';
}

1;