package XServSubserver;

use IO::Socket;
use HTTP::Daemon;
use HTTP::Status;
use XML::Parser;
use XML::Writer;
use POSIX ":sys_wait_h";
use File::Temp qw(tempfile);
use MIME::Entity;
# use Devel::Size qw(size total_size);

# use Data::Dumper;

my $ErrCode = 0;

my $EOL = "\n";
my $END_OF_SEND = "<xcom_end>";
my $HTTP_PREFIX;
# my $MIME = 0;
# my @CONN = ();
# my $C;
my @XSubKids;

sub main
{
  if (defined($::ts_useSSL))
  {
    require IO::Socket::SSL;
    require HTTP::Daemon::SSL;
    $HTTP_PREFIX = "https://";
  }
  else { $HTTP_PREFIX = "http://" }

  if ( ($::ts_ssTempFiles == 1) || ( defined($::ts_prSock) && (($::ts_taskAPI eq 'Files') || ($::ts_taskAPI eq 'Proxy')) ) )
  {
    if (! -d $::ts_ssTempDir)
    { mkdir ($::ts_ssTempDir) or die ("XServSubServer: can't create $::ts_ssTempDir!\n") }
    unlink <$::ts_ssTempDir/*>;
  }

  my $SubServer = HTTP::Daemon->new(LocalPort => $::ts_ssPort, Type => SOCK_STREAM, ReuseAddr => 1, Listen => SOMAXCONN, Proto => "tcp")
    or die ("XServSubServer: can't create socket for incoming clients: $!\n");
  my @DataFromXServ=();
  my @Data4XServ=();

  if ($::ts_taskAPI eq 'Subserver')
  {
    my $XServSocket = connectXServ('INI');
    print $XServSocket "INI".$EOL x 2;
    while (defined(my $line=<$XServSocket>))
    {
      chomp $line;
      push (@DataFromXServ, $line);
    }
    close $XServSocket;
    ($::ts_taskName, $::ts_taskId, $::ts_taskArgs) = @DataFromXServ;
    print STDERR "\n$::ts_taskName :: $::ts_taskId :: $::ts_taskArgs\n";
    XServUtils::LogWrite2("\t",'INIT',"TaskName=$::ts_taskName&TaskId=$::ts_taskId&TaskArgs={$::ts_taskArgs}");
    print STDERR "XServSubserver: initialized\n";
  }
  my $pid_ss;
  while (my $Client = $SubServer->accept())
  {
    # next if my $pid_ss=fork();
    $pid_ss=fork();
    while (!defined ($pid_ss))
    {
      # die ("XServSubServer: can't fork: $!\n") if !defined($pid_ss);
      print STDERR "XServSubServer: can't fork: $!\n";
      $pid_ss=fork();
    }
    next if $pid_ss;
    my $tStart = Time::HiRes::time();
    my ($prepStart, $prepEnd);
    $SubServer->close() if ($^O eq 'MSWin32');
    # binmode $Client;
    if (defined($::ts_useSSL) && ($::ts_useSSL=~/^verify$/ix))
    {
      $Client = HTTP::Daemon::ClientConn::SSL->start_SSL(
      $Client,
      SSL_ca_path => $::ts_SSL_trusted_dir,
      SSL_verify_mode =>0x3,
      SSL_key_file => $::ts_SSL_key_file,
      SSL_cert_file => $::ts_SSL_cert_file,
      SSL_server =>1 );
    }
    if (defined($::ts_useSSL) && ($::ts_useSSL=~/^verifycrl$/ix))
    {
      $Client = HTTP::Daemon::ClientConn::SSL->start_SSL(
      $Client,
      SSL_ca_path => $::ts_SSL_trusted_dir,
      SSL_verify_mode =>0x3,
      SSL_key_file => $::ts_SSL_key_file,
      SSL_cert_file => $::ts_SSL_cert_file,
      SSL_server =>1,
      SSL_check_crl=>1 ); # must use OpenSSL>0.9.7 or it will cause an error
    }
    elsif (defined($::ts_useSSL) && ($::ts_useSSL=~/^noverify$/ix))
    {
      $Client = HTTP::Daemon::ClientConn::SSL->start_SSL(
      $Client,
      SSL_ca_path => $::ts_SSL_trusted_dir,
      SSL_verify_mode =>0x2,
      SSL_key_file => $::ts_SSL_key_file,
      SSL_cert_file => $::ts_SSL_cert_file,
      SSL_server =>1 );
    }
    else
    {
      # binmode $Client;
      # print STDERR "mydebug:noSSL\n";
    }
    binmode $Client;

    # my $ClRequest = ( $::ts_ssReqOnDisk == 1 ? $Client->get_request(1) : $Client->get_request() );
    # my $ClRequestHeaders = $Client->get_request(1);
#print STDERR "[1]\n";
    my $ClRequest = $Client->get_request();
#print STDERR "[2]\n";
# print STDERR $ClRequest->as_string();
    if (!defined($ClRequest))
    {
      $ErrCode = 1;
      print STDERR "S> Incomplete or broken request\n";
      exit;
    }
    @DataFromXServ=();
    @Data4XServ=();
    my @logPList=(); my $logPTime=0; my $logSess; my $logPSize=0;
    my $XMLData4Client;
    my $XMLWriter = new XML::Writer (OUTPUT => \$XMLData4Client);

    if ($ClRequest->method eq 'POST')
    {
      my $p = new XML::Parser(Style => 'Objects');
      my ($RequestType, $NodeId, $NodeOS, $NodeCPU, $NodeCPUNumCores, $NodeCPUFreq, $NodeCPUPerf, $NodeRAMSize, $TaskName, $TaskId, $SessionId);
#print STDERR "[3]\n";
      my @ClRequestParts=$ClRequest->parts;
#print STDERR "[4]\n";
      my $FirstPart = (@ClRequestParts == 0 ? $ClRequest->decoded_content() : $ClRequestParts[0]->decoded_content());
#print STDERR "[5]\n";
      my $RequestParsed = $p->parse($FirstPart);
      my $Response4Client = HTTP::Response->new(200);
      my $MIMEResponse4Client;
      # my ($genPortionTime, $addPortionTime) = (0, 0);
      my ($APITime, $IntCommTime) = (0, 0);
      my ($IntCommStart, $IntCommEnd);
      $prepStart = Time::HiRes::time();
      if ($RequestParsed->[0]->isa('XServSubserver::get-task'))	# TSR, get-task, task request
      {
	push (@Data4XServ, $RequestType='TSR',
	  $Client->peerhost(),
	  $NodeId=$RequestParsed->[0]->{Kids}->[0]->{id},
	  $NodeCPU=$RequestParsed->[0]->{Kids}->[0]->{Kids}->[0]->{type},
          $NodeCPUNumCores=$RequestParsed->[0]->{Kids}->[0]->{Kids}->[0]->{'num-cores'},
	  $NodeCPUFreq=$RequestParsed->[0]->{Kids}->[0]->{Kids}->[0]->{freq},
	  $NodeCPUPerf=$RequestParsed->[0]->{Kids}->[0]->{Kids}->[0]->{perf},
	  $NodeRAMSize=$RequestParsed->[0]->{Kids}->[0]->{Kids}->[1]->{size},
	  $NodeOS=$RequestParsed->[0]->{Kids}->[0]->{Kids}->[2]->{type});
	$IntCommStart = Time::HiRes::time();
	my $XServSocket = connectXServ('TSR');
	print $XServSocket join($EOL,@Data4XServ).$EOL x 2;
	# print $XServSocket join($EOL,@Data4XServ).$EOL;
	while (defined(my $line=<$XServSocket>))
	{ 
          chomp $line;
	  if ($line eq 'ERR')
	  {
	    $ErrCode = 1;
	    print STDERR "S[TSR]> ERR received from XServProcessor\n";
	    last;
	  }
	  push (@DataFromXServ, $line);
	}
	close $XServSocket;
	$IntCommEnd = Time::HiRes::time();
	$IntCommTime = $IntCommEnd - $IntCommStart;
	$XMLWriter->startTag('process-task', 'session-id'=>($logSess=$DataFromXServ[0]), 'api'=>$::ts_xCliAPI, 'task-name'=>$::ts_taskName, 'task-id'=>$::ts_taskId, 'task-args'=>$::ts_taskArgs, (($::ts_xCliAPI eq 'Processes') && (defined($::ts_xUser)) ? ('user'=>$::ts_xUser) : undef));
	my $execFileName=XServUtils::GetTaskFilename($::ts_taskName,$NodeOS,$NodeCPU);
	$XMLWriter->startTag('file', 'executable'=>'false', 'archive'=>'true', 'filename'=>$execFileName, 'id'=>'', 'order'=>$::ts_dlOrder);
	foreach my $dlSite (@::ts_dlSites) { $XMLWriter->dataElement('url',$dlSite.$execFileName) }
	$XMLWriter->endTag('file');
	foreach my $dlFile (keys %::ts_dlFiles)
	{
	  $XMLWriter->startTag('file', 'executable'=>'false', 'filename'=>$dlFile, 'id'=>'', 'order'=>$::ts_dlOrder);
	  foreach my $dlSite (@::ts_dlSites) { $XMLWriter->dataElement('url',$dlSite.$dlFile) }
	  $XMLWriter->endTag('file');
	}
	$XMLWriter->startTag('portion-server','order'=>'fallback');
	$XMLWriter->dataElement('url',$HTTP_PREFIX.$::ts_ssREQHost.':'.$::ts_ssREQPort);
	$XMLWriter->endTag('portion-server');
	$XMLWriter->startTag('result-server','order'=>'fallback');
	$XMLWriter->dataElement('url',$HTTP_PREFIX.$::ts_ssASWHost.':'.$::ts_ssASWPort);
	$XMLWriter->endTag('result-server');
	$XMLWriter->startTag('status-server','order'=>'fallback','interval'=>'300');
	$XMLWriter->dataElement('url',$HTTP_PREFIX.$::ts_ssHost.':'.$::ts_ssPort);
	$XMLWriter->endTag('status-server');

	if ($::ts_xCliAPI eq 'Processes')
	{
	  $XMLWriter->dataElement('init-command', $::ts_xCliInitCmd);
	  $XMLWriter->dataElement('portion-command', $::ts_xCliPortionCmd, 'portion'=>$::ts_xCliPortionIn, 'result'=>$::ts_xCliPortionOut);
	}

	$XMLWriter->endTag('process-task');
	$XMLWriter->end();
	$Response4Client->content($XMLData4Client);
	print STDERR $XMLData4Client if ($::ts_debugShowXML == 1);
      }
      elsif ($RequestParsed->[0]->isa('XServSubserver::get-portion'))
      {
	my ($pNumber, $pContent, $pSize, $pArgs, $pFile, $tempFile, $tempFileName);
	my @mParts=();
	my @pFiles=();
	if (($RequestParsed->[0]->{'task-name'} eq $::ts_taskName) && ($RequestParsed->[0]->{'task-id'} eq $::ts_taskId))
	{
	  $NodeId=$RequestParsed->[0]->{'client-id'};
  	  push (@Data4XServ, $RequestType='REQ',
            $logSess=$SessionId=$RequestParsed->[0]->{'session-id'});	# Trying to pass only SessId to the Processor
	  my $q = (defined($RequestParsed->[0]->{'q'}) ? $RequestParsed->[0]->{'q'} : 1);
	  $XMLWriter->startTag('portions', 'q' => $q);
	  while ($q>0)
	  {
	    my $ln = 0; my ($l1, $l2); $pContent = '';
	    if ($::ts_ssTempFiles != 0)
	    {
	      ($tempFile, $tempFileName) = tempfile (DIR => $::ts_ssTempDir, UNLINK => 1)
	        or die ("XServSubserver(REQ): can't create temporary file in $::ts_ssTempDir!\n");
	      binmode $tempFile;
	    }
	    $IntCommStart = Time::HiRes::time();
	    my $XServSocket = connectXServ('REQ');
	    print $XServSocket join($EOL,@Data4XServ).$EOL x 2;
	    while (defined(my $line=<$XServSocket>))
	    {
# print STDERR 'SS>>'.$line;
	      if ($ln == 0) { $l1 = $line; chomp $l1; }
	      elsif ($ln == 1) { $l2 = $line; chomp $l2; }
	      elsif ($ln == 2) { $l3 = $line; chomp $l3; }
	      else
	      { 
		if ( defined($::ts_prSock) && (($::ts_taskAPI eq 'Files') || ($::ts_taskAPI eq 'Proxy')) )
		{ $pFile = $line; chomp $pFile }
		elsif ($::ts_ssTempFiles != 0)
		{ print $tempFile $line or die ("XServSubserver(REQ): can't write to $::ts_ssTempDir/$tempFileName!\n") }
		else
		{ $pContent .= $line }
	      }
	      $ln++;
	    }
	    close $XServSocket;
	    $IntCommEnd = Time::HiRes::time();
	    $IntCommTime += $IntCommEnd - $IntCommStart;
	    if ($l1 eq 'ERR')
	    {
	      $ErrCode = 1;
	      print STDERR "S[REQ]> ERR received from XServProcessor\n";
	      last;
	    }
	    else
	    {
	      $pNumber = $l1; $pArgs = $l2; $APITime = $l3;
	      if ($::ts_ssTempFiles != 0)
	      {
		close $tempFile;
		$pFile = $tempFileName;
	      }
	    }

	    push(@logPList, $pNumber);
	    $pSize = (defined $pFile ? (stat($pFile))[7] : length ($pContent));
	    $logPSize += $pSize;
	    $XMLWriter->emptyTag('portion', 'num' => $pNumber, 'size' => $pSize, 'args' => $pArgs);
	    if (defined ($pFile))
	    {
	      if (-f $pFile) { push (@pFiles, $pFile) } else { die ("XServSubServer(REQ): can't find $pFile!\n") }
	    }
	    else
	    {
	      push (@mParts, HTTP::Message->new(HTTP::Headers->new(Content_Type => 'application/octet-stream'),$pContent));
	    }
	    @DataFromXServ=();
	    $q--;
	  }
	  if ($ErrCode == 0)
	  {
	    $XMLWriter->endTag('portions');
            $XMLWriter->end();
	    if (@pFiles == 0)
	    {
	      unshift (@mParts, HTTP::Message->new(HTTP::Headers->new(Content_Type => 'text/xml'), $XMLData4Client));
	      $Response4Client->parts(\@mParts);
	    }
	    else
	    {
	      $MIMEResponse4Client = MIME::Entity->build(Type => 'multipart/form-data');
	      my $mime_part = MIME::Entity->build(Type => 'text/xml', Data => $XMLData4Client, 'Content-Disposition:' => 'form-data');
	      $mime_part->head()->mime_attr('content-type.name' => 1);
	      $MIMEResponse4Client->add_part($mime_part);
	      my $i=2;
	      foreach my $pf (@pFiles)
	      {
	        $mime_part = MIME::Entity->build(Type => 'application/octet-stream', Path => $pf, 'Content-Disposition:' => 'form-data', Encoding => 'binary');
	        $mime_part->head()->mime_attr('content-type.name' => $i);
	        $MIMEResponse4Client->add_part($mime_part);
	        $i++;
	      }
	      # my $counter = IO::Counter->new();
	      # $MIMEResponse4Client->print_body($counter);
	      # my $len = $counter->counter();
	      # $MIMEResponse4Client->head()->replace('Content-Length', $len);
	      $MIMEResponse4Client->sync_headers(Length=>'COMPUTE');
	    }
	  }
	}
	else 
	{ 
	  $ErrCode = 1;
	  print STDERR "S[REQ]> Client requested portion for incorrect task\n";
	}
      }
      elsif ($RequestParsed->[0]->isa('XServSubserver::post-result'))
      {
	$RequestType='ASW';
	if (($RequestParsed->[0]->{'task-name'} eq $::ts_taskName) && ($RequestParsed->[0]->{'task-id'} eq $::ts_taskId))
	{
	  $NodeId=$RequestParsed->[0]->{'client-id'};
	  my $q = $RequestParsed->[0]->{'q'};
	  push (@Data4XServ, $RequestType='ASW', 
	    $logSess=$SessionId=$RequestParsed->[0]->{'session-id'});
	  for (my $i=1; $i<=$q; $i++)
	  {
	    my ($pNum, $pSize, $pTime) = ($RequestParsed->[0]->{Kids}->[$i-1]->{'num'}, $RequestParsed->[0]->{Kids}->[$i-1]->{'size'}, $RequestParsed->[0]->{Kids}->[$i-1]->{'time'});
	    my $td = $ClRequestParts[$i]->decoded_content();
	    my $lentd = length($td);
	    $logPSize += $lentd;
	    if ($pSize != $lentd)
	    {
	      $ErrCode = 1;
	      print STDERR "S[ASW]> Unexpected portion size\n";
	      last;
	    }
	    push (@logPList, $pNum); $logPTime+=$pTime;
	    if ( defined($::ts_prSock) && (($::ts_taskAPI eq 'Files') || ($::ts_taskAPI eq 'Proxy')) )
	    {
	      my ($tempFile, $tempFileName) = tempfile (DIR => $::ts_ssTempDir, UNLINK => 0) 
	        or die ("XServSubserver(ASW): can't create temporary file in $::ts_ssTempDir!\n");
	      binmode $tempFile;
	      print $tempFile $td
	        or die ("XServSubserver(ASW): can't write to $::ts_ssTempDir/$tempFileName!\n");
	      close $tempFile;
	      $td = $tempFileName;
	    }
	    $IntCommStart = Time::HiRes::time();
	    my $XServSocket = connectXServ('ASW');
	    print $XServSocket join($EOL,@Data4XServ).$EOL.$pNum.$EOL.$pTime.$EOL.$td.$EOL.$END_OF_SEND.$EOL;
	    my $line;
	    while (defined($line=<$XServSocket>))
	    {
#print STDERR "S> $line";
	      chomp $line;
	      if ($line eq 'ERR') 
	      { 
	        $ErrCode=1; 
		print STDERR "S[ASW]> ERR received from XServProcessor\n";
	        last;
	      }
	      else 
	      { 
	        $APITime = $line if $line ne '';
	      }
	    }
	    close $XServSocket;
	    $IntCommEnd = Time::HiRes::time();
	    $IntCommTime += $IntCommEnd - $IntCommStart;
	    last if ($ErrCode != 0);
	  }
	}
	else 
	{ 
	  $ErrCode = 1;
	  print STDERR "S> Unknown or incorrect request\n";
	}
      }
      if ($ErrCode != 0)
      {
	$RequestType = 'INIT';
	my $XMLError4Client;
        my $XMLErrorWriter = new XML::Writer (OUTPUT => \$XMLError4Client);
        $XMLErrorWriter->emptyTag('initialize');
        $XMLErrorWriter->end();
	$Response4Client->content($XMLError4Client);
      }
#print STDERR "A";
	
#print STDERR "<".total_size($Response4Client).">";
      $prepEnd = Time::HiRes::time();
      if (!defined($MIMEResponse4Client))
      { 
        my $cr = $Client->send_response($Response4Client);
      } # if ($Response4Client->content ne '')
      else
      {
#open FFF, "> /tmp/ss_resp.txt";
#$MIMEResponse4Client->print(\*FFF);
#close FFF;
        $Client->send_status_line(200);
        $MIMEResponse4Client->print($Client);
      }
#print STDERR "B";
      my $tEnd = Time::HiRes::time();
      if (!$::ts_logOldFormat)
      {
	XServUtils::LogWrite2($Client->peerhost(),$RequestType,'SId='.$logSess.'&Node='.$NodeId.'&STime='.$tStart.'&prepStart='.$prepStart.'&intCommTime='.$IntCommTime.'&prepEnd='.$prepEnd.'&ETime='.$tEnd.($RequestType eq 'ASW' ? "&Time=$logPTime" : '').($RequestType ne 'TSR' ? "&APITime=$APITime&Portion={".join(',',@logPList)."}&Length=$logPSize" : "&OS=$NodeOS&CPUType=$NodeCPU&CPUNumCores=$NodeCPUNumCores&CPUFreq=$NodeCPUFreq&CPUPerf=$NodeCPUPerf&RAMSize=$NodeRAMSize"));
      }
      elsif ($RequestType ne 'INIT')
      {
        XServUtils::LogWrite($Client->peerhost(),$RequestType.'1:sId='.$logSess.'&gcCode='.$NodeId.'&stime='.$tStart);
	XServUtils::LogWrite($Client->peerhost(),$RequestType.'2:sId='.$logSess.'&gcCode='.$NodeId.'&stime='.$tEnd.($RequestType eq 'ASW' ? "&time=".$logPTime : '').($RequestType ne 'TSR' ? '&portion='.($#logPList==0 ? $logPList[0] : '{'.join(',',@logPList).'}').'&length='.$logPSize : ''));
	
      }
      $ErrCode = 0 if $ErrCode != 0;
#print STDERR "C";
    }
    elsif ($ClRequest->method eq 'GET')
    {
      my $FileRequested = substr($ClRequest->url->path, 1);
      if ($FileRequested eq '' || $FileRequested eq 'html' || $FileRequested eq 'status.html')
      { $FileRequested = 'HTML' }
      elsif ( $FileRequested eq 'xml' || $FileRequested eq 'status.xml' )
      { $FileRequested = 'XML' }
      my %Form = $ClRequest->url->query_form;
      if (($FileRequested eq 'HTML') || ($FileRequested eq 'XML'))	# Status request
      {
	my $XServSocket = connectXServ('STAT');
	print $XServSocket "STA".$EOL x 2;
	while (defined(my $line=<$XServSocket>))
	{
#print STDERR 'SS>>'.$line;
	  chomp $line;
	  if ($line eq '') { last };
	  push (@DataFromXServ, $line);
	}
	close $XServSocket;
	my $MLStat = XServUtils::GenerateStatML (\@DataFromXServ, $FileRequested, (defined $Form{'r'} ? $Form{'r'} : 0));
	my $Response4Client = HTTP::Response->new(200);
	$Response4Client->content($MLStat);
	$Client->send_response($Response4Client);
	my $tEnd = Time::HiRes::time();
	if (!$::ts_logOldFormat)
	{
	  XServUtils::LogWrite2($Client->peerhost(),'STAT','STime='.$tStart.'&ETime='.$tEnd);
	}
      }
      elsif (exists($::ts_dlFiles{$FileRequested}))	# Data file request, look only in specified path
      {
	my $fn = $::ts_dlFiles{$FileRequested}.'/'.$FileRequested;
	$Client->send_file_response($fn);
	my $fsz = (stat($fn))[7];
	my $tEnd = Time::HiRes::time();
	if (!$::ts_logOldFormat)
	{
	  XServUtils::LogWrite2($Client->peerhost(),'DFILE',(defined($Form{'client-id'})?'Node='.$Form{'client-id'}.'&':'').'File='.$FileRequested.'&STime='.$tStart.'&ETime='.$tEnd.'&Length='.$fsz);
	}
      }
      else	# Task file request
      {
	$FileRequested =~ s/\.+/\./g;
	my $FileRPath;
	if ($::ts_taskAPI ne 'Proxy') { $FileRPath = $::ts_taskPath.'/'.$FileRequested }
	else { $FileRPath = "$::ts_proxyTmp/$::ts_taskName.$::ts_taskId.$::ts_sessId/$FileRequested" }
# print STDERR $FileRPath;
	if ( -f $FileRPath )
	{
	  $Client->send_file_response($FileRPath);
	  my $fsz = (stat($FileRPath))[7];
	  my $tEnd = Time::HiRes::time();
	  if (!$::ts_logOldFormat)
	  {
	    XServUtils::LogWrite2($Client->peerhost(),'TFILE',(defined($Form{'client-id'})?'Node='.$Form{'client-id'}.'&':'').'File='.$FileRequested.'&STime='.$tStart.'&ETime='.$tEnd.'&Length='.(defined $fsz ? $fsz : 0));
	  }
	}
	else
	{
	  my $Response4Client = HTTP::Response->new(404);
	  $Client->send_response($Response4Client);
	}
      }
    }
    exit();
  } 
  continue 
  {
    $Client->close();
    if ($^O eq 'MSWin32')	# This is not necessary under Linux
    {
      push (@XSubKids, $pid_ss);
      my @kids;
      foreach my $childpid (@XSubKids) 
      {
        if (waitpid ($childpid, WNOHANG) != $childpid) { push (@kids, $childpid) }
      }
      @XSubKids=@kids;
    }
  }
}

sub connectXServ
{
  my $reason = $_[0];
  my $tss;
  my $cnt = 1;
  # my $time0 = time();
  while ($cnt <=20)
  {
    if (!defined($::ts_prSock))
    { $tss = IO::Socket::INET->new(PeerAddr => $::ts_prHost, PeerPort => $::ts_prPort, Proto => "tcp", Type => SOCK_STREAM, Timeout => 1) }
    else
    { $tss = IO::Socket::UNIX->new(Type => SOCK_STREAM, Peer => $::ts_prSock) }
    # or die ("S> Can't connect to Proccessor at $::ts_prHost:$::ts_prPort");
    last if $tss;
    print STDERR "XServSubserver: can't connect to ".(defined($::ts_prSock)?$::ts_prSock:"$::ts_prHost:$::ts_prPort")." ($reason), retrying ($cnt)...\n";
    sleep (1); 
    $cnt++ if $::ts_taskAPI ne 'Subserver';
  }
  # my $time1 = time();
  if ($tss) 
  {
    binmode $tss;
# print STDERR "S> $reason\n";
    return $tss;
  }
  else
  {
    die ("XServSubserver: Giving up, sorry!\n")
  }
}

1;