#!/usr/bin/perl
#
#	-newca
#	-sign file1 file2 ... takes certificate request 'file1' as input and creates a certificate 'file2' as output
#       -newcert file1 file2 ... creates a key (file1) and certificate (file2) for client
#         

my $openssl;
if(defined $ENV{OPENSSL}) {
	$openssl = $ENV{OPENSSL};
} else {
	$openssl = "openssl";
	$ENV{OPENSSL} = $openssl;
}

$SSLEAY_CONFIG=$ENV{"SSLEAY_CONFIG"};
$DAYS="-days 365";	# 1 year
$CADAYS="-days 1095";	# 3 years
$REQ="$openssl req $SSLEAY_CONFIG";
$CA="$openssl ca $SSLEAY_CONFIG";
$VERIFY="$openssl verify";
$X509="$openssl x509";
$PKCS12="$openssl pkcs12";


$CATOP="./demoCA";
$CAKEY="cakey.pem";
$CAREQ="careq.pem";
$CACERT="cacert.pem";

$DIRMODE = 0777;

$RET = 0;

if (@ARGV) {
	if ( $ARGV[0] =~ /^-newca$/ ) {
	    print "Setting things up...\n";
	   $NEW="1";
	    if ( "$NEW" || ! -f "${CATOP}/serial" ) {
		# create the directory hierarchy
		mkdir $CATOP, $DIRMODE;
		mkdir "${CATOP}/certs", $DIRMODE;
		mkdir "${CATOP}/crl", $DIRMODE ;
		mkdir "${CATOP}/newcerts", $DIRMODE;
		mkdir "${CATOP}/private", $DIRMODE;
		open OUT, ">${CATOP}/index.txt";
		close OUT;
		open OUT, ">${CATOP}/serial";
		print OUT "F22F02B849CA1288\n";
		close OUT;
		open OUT, ">${CATOP}/crlnumber";
		print OUT "01\n";
		close OUT;
	    }
	    if ( ! -f "${CATOP}/$CACERT" ) {
		print "CA certificate filename (or enter to create a request)\n";
		$FILE = <STDIN>;
		chop $FILE;
		# ask user for existing CA certificate
		if ($FILE) {
		    cp_pem($FILE,"${CATOP}/private/$CAKEY", "PRIVATE");
		    cp_pem($FILE,"${CATOP}/$CACERT", "CERTIFICATE");
		    $RET=$?;
		} else {
		    print "Making a request ...\n";
		    system ("$REQ -new -keyout " .
			"${CATOP}/private/$CAKEY -out ${CATOP}/$CAREQ");
		    $RET=$?;
		    print "All you need now to get started with issuing your own\n".
		    "client certificates is to sign a certificate for your CA.\n".
		    "Ask your root certification authority for details.\n";
		}
	    }
	} elsif ($ARGV[0] =~ /^-newcert$/) {
	    #create a certificate
	    print "Making a client keys+certificate..\n";
	    $outkey = "newkey.pem";
	    $outcert = "newcert.pem";
	    $outreq = "tmpreq.pem";
	    if ($ARGV[1] && $ARGV[2]) {
		$outkey = $ARGV[1];
		$outcert = $ARGV[2];
	    }
		system ("$REQ -new -keyout $outkey -out $outreq $DAYS");
	    system ("$CA -policy policy_anything -out $outcert " .
							"-infiles $outreq");
	    $RET=$?;
	    print "Certificate is in $outcert, private key is in $outkey\n";
	} elsif ($ARGV[0] =~ /^-sign$/) {
	    #make a certificate from request
	    print "Signing a certificate...\n";
	    $outcert = "newcert.pem";
	    $inreq = "newreq.pem";
	    if ($ARGV[1]) {
		$inreq = $ARGV[1];
		if ($ARGV[2])
		{
		    $outcert = $ARGV[2];
		}
	    }
	    system("$CA -policy policy_anything -out $outcert " . 
							"-infiles $inreq");
	    print "Certificate is in $outcert\n";
	    
	}elsif ($ARGV[0] =~ /^-verify$/) {
	     $incert = "newcert.pem";
	     if ($ARGV[1]) {
		$incert = $ARGV[1];
	     }    
	     print "Verifying a certificate $incert...\n";
	     print ("$VERIFY -verbose -CAfile $CATOP/$CACERT $incert\n");
	     system ("$VERIFY -verbose -CAfile $CATOP/$CACERT $incert");
	     $RET=$?;
	     print "$RET\n";
	}elsif (($ARGV[0] =~ /^-revoke$/) && ($ARGV[1])) {
	     $incert = $ARGV[1];
	     print "Revoking a certificate $incert...\n";
	     print ("$CA -revoke $incert\n");
	     system ("$CA -revoke $incert");
	     $RET=$?;
	     print "$RET\n";
	}elsif($ARGV[0] =~ /^-gencrl$/) {
		 $outcert = "crl.pem";
		 if ($ARGV[1]) {
		 $outcert = $ARGV[1];
		 }
		 system ("$CA -gencrl -out $outcert");
		 $RET=$?;
		 print "CRL is in $outcert\n";
		 print "$RET\n";
	}
 }  else {
 print STDERR "usage: CA -newcert [outkeyfile, outcertfile]|-newca|".
					"-sign [inreqfile outcertfile]|".
					"-revoke incertfile|".
					"-gencrl [crlfile]\n";
}
	

exit $RET;


sub cp_pem {
    my ($infile, $outfile, $bound) = @_;
    open IN, $infile;
    open OUT, ">$outfile";
    my $flag = 0;
    while (<IN>) {
	$flag = 1 if (/^-----BEGIN.*$bound/) ;
	print OUT $_ if ($flag);
	if (/^-----END.*$bound/) {
	    close IN;
	    close OUT;
	    return;
	}
    }
}

1;
