package Pi;

use strict;
use warnings;
use POSIX qw(locale_h);

use vars qw (@ISA @EXPORT);
use Exporter;
@ISA = qw (Exporter);
@EXPORT = qw (initialize getFirstPortionNumber getLastPortionNumber isFinished getPortion addPortion finalize);

my $totalSum;			# The sum of drops which have fallen inside of circle
my $donePortions;		# Done portions count
my $expo;			# Power of 10, determines the number of drops inside square (0,0)-(1,1) per portion
my $portionsCount;		# Number of portions
my $thePI;			# Resulting Pi number
my $outCSV;			# Output file for CSV data
my $outRes;			# Output file for Pi value
my $PI = 3.141592653589793238;	# Real Pi (for comparison)

sub initialize {	
	($portionsCount,$expo) = split (/ /, $_[0])
		or die ("Pi error: incorrect arguments!");	#Setting up task arguments
	$totalSum = 0;
	$donePortions = 0;
	$outCSV = "../tasks/Pi/out/pi_esteem_$portionsCount"."_$expo.csv";
	$outRes = "../tasks/Pi/out/pi_$portionsCount"."_$expo.txt";
	open E, "> $outCSV"
		or die ("Pi error: can't write to $outCSV!");
	print E 'Portions;Drops;EstimatedPi;RealPi;Error'."\n";
	close E;
	print STDERR "Pi initialized: $portionsCount portions, ".(10**$expo)." (10**$expo) drops in portion\n";
	return 1;
}

sub getFirstPortionNumber {
	return 1;		# First portion is 1
}

sub getLastPortionNumber {
	return $portionsCount;	# Last portion is $portionsCount
}

sub getPortion {
	my ($portion) = @_;
	return $portion;	# Returning portion number as portion body
}

sub addPortion {
	my ($n, $portionSum) = ($_[0], $_[1]);
	$totalSum = $totalSum + $portionSum;
	$donePortions += 1;
	my $dropsCount = $donePortions * (10 ** $expo);
	$thePI = $totalSum / $dropsCount;
	open E, ">> $outCSV"
		or die ("Pi error: can't write to $outCSV!");
	my $lc_numeric = setlocale(LC_NUMERIC);
	print E "$donePortions;$dropsCount;$thePI;$PI;".($thePI-$PI)."\n";
	setlocale(LC_NUMERIC,$lc_numeric);
	close E;
}

sub isFinished {
	return 0;		# Knowing $portionsCount, should always return 0
}

sub finalize {
	open F, "> $outRes"
		or die ("Pi error: can't write to $outRes");
	print F $thePI."\n";
	close F;
	return 1;
}

1;

