package TestPl;

use strict;
use warnings;
use POSIX 'strftime';

use vars qw (@ISA @EXPORT);
use Exporter;
@ISA = qw (Exporter);
@EXPORT = qw (initialize getFirstPortionNumber getLastPortionNumber isFinished getPortion addPortion finalize);

my ($pOutSizeMin,$pOutSizeMax);	# min & max outgoing (srv->clnt) portion sizes, bytes
my $pOutStrategy;		# distribution strategy for outgoing portion sizes (C,R,I,D)
my ($pInSizeMin,$pInSizeMax);	# min & max incoming (clnt->srv) portion sizes, bytes
my $pInStrategy;		# distribution strategy for incoming portion sizes (C,R,I,D)
my ($cTimeMin,$cTimeMax);	# computation time for client, seconds
my $cTimeStrategy;		# distribution strategy for client times (C,R,I,D,E)
my $mSize;			# matrix size for 'matrix', 'matrix_light' or 'matrix_ultralight', if needed
my $mProg;			# (N,M,L,U)
my $pCount;			# total number of portions

my @s = qw (1 2 3 4 5 6 7 8 9 0);

sub initialize
{
  if ($_[0] =~ /(\d+) (\d+) ([CRID]) (\d+) (\d+) ([CRID]) (\d+) (\d+) ([CRIDE]) (\d+) ([MLUN]) (\d+)/i)
  { ($pOutSizeMin, $pOutSizeMax, $pOutStrategy, $pInSizeMin, $pInSizeMax, $pInStrategy, $cTimeMin, $cTimeMax, $cTimeStrategy, $mSize, $mProg, $pCount) = ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12) }
  else
  { die ("TestPl error: incorrect arguments!") }
  print STDERR "TestPl arguments:\npOutSizeMin\t = $pOutSizeMin\npOutSizeMax\t = $pOutSizeMax\npOutStrategy\t = $pOutStrategy\npInSizeMin\t = $pInSizeMin\npInSizeMax\t = $pInSizeMax\npInStrategy\t = $pInStrategy\ncTimeMin\t = $cTimeMin\ncTimeMax\t = $cTimeMax\ncTimeStrategy\t = $cTimeStrategy\nmSize\t\t = $mSize\nmProg\t\t = $mProg\npCount\t\t = $pCount\n";
  unlink '../tasks/TestPl/portions.in', '../tasks/TestPl/portions.out';
  return 1;
}

sub getFirstPortionNumber
{
  return 1;
}

sub getLastPortionNumber
{
  return $pCount;
}

sub isFinished
{
  return 0;
}

sub getPortion
{
  my $n = $_[0];
  my $pLength;
  my $portion = '';
  if ($pOutStrategy eq 'R') 		#R for Random
  { $pLength = $pOutSizeMin+(int(rand($pOutSizeMax-$pOutSizeMin+1))); }
  elsif ($pOutStrategy eq 'I')		# I for Increase
  { $pLength = $pOutSizeMin+int(($pOutSizeMax-$pOutSizeMin)*$n/$pCount); }
  elsif ($pOutStrategy eq 'D')		# D for Decrease
  { $pLength = $pOutSizeMax-int(($pOutSizeMax-$pOutSizeMin)*$n/$pCount); }
  else 					# C for Constant
  { $pLength = $pOutSizeMin; }
  my $a = 1;
  for (my $i=0; $i<$pLength; $i++)
  {
    $portion.=$s[$i%10];
  }
  open F, ">> ../tasks/TestPl/portions.out"
    or die ("TestPl error: can't write to portions.out!");
  print F "$n\t$pLength\n";
  close F;
  return $portion;
}

sub addPortion
{
  my ($n, $data) = ($_[0], $_[1]);
  open F, ">> ../tasks/TestPl/in/result.dat.".strftime('%Y-%m-%d-%H', localtime)
    or die ("TestPl error: can't write to result.dat!");
  print F $data."\n";
  close F;
  open F, ">> ../tasks/TestPl/portions.in"
    or die ("TestPl error: can't write to portions.in!");
  print F $n."\t".length($data)."\n";
  close F;
}

sub finalize
{
  print STDERR "TestPl finished.\n";
}

1;