#!/usr/bin/perl -W

use POSIX qw (mktime);
use Time::localtime;
use POSIX qw(locale_h);
# use Data::Dumper;
use strict;
no strict "refs";
use warnings;

sub mkt  # 2008-12-19 11-08-36.912486 -> UNIX timestamp
{ 
  my $s = $_[0];
  return mktime (substr($s,17,2 ),substr($s,14,2),substr($s,11,2),substr($s,8,2),substr($s,5,2)-1,substr($s,0,4)-1900).substr($s,19);
}

sub mt  # sec -> hh:mm:ss
{
  my $sec = $_[0];
  my $th = int($sec/3600);
  my $tm = int(($sec-$th*60*60)/60);
  my $ts = $sec-$tm*60-$th*60*60;
  return sprintf("%02d:%02d:%02d",$th,$tm,$ts);
}

sub lct # UNIX timestamp -> 2008-12-19 11-08-36
{
  my $t = localtime($_[0]);
  return (1900+$t->year()).'-'.z(1+$t->mon()).'-'.z($t->mday()).' '.z($t->hour()).':'.z($t->min()).':'.z($t->sec());
}

sub ifz { return (defined $_[0] ? $_[0] : 0) }

sub z { return ($_[0] > 9 ? $_[0] : '0'.$_[0]) }

sub getpt
{
  my $plist = $_[0];
  $plist = substr ($plist, 1, length($plist)-2);
  return split /,/, $plist;
}

sub min
{
  my $min = $_[0];
  foreach my $val (@_)
  { $min = $val if ($val < $min) }
  return $min;
}

sub max
{
  my $max = $_[0];
  foreach my $val (@_)
  { $max = $val if ($val > $max) }
  return $max;
}

die ("Usage: $0 <log_file_name> [options]
Available options:
 -i <seconds>\tinterval (in seconds) of table data for graphics building
 -t <template>\ttemplate file for general information
 -m <new|add>\tfile writing mode (create new or append to existing file)
 -o <filename>\tout file name for general info
 -ot <filename>\tout file name for general table
 -op <filename>\tout file name for portions table
 -on <filename>\tout file name for nodes data table
") if @ARGV < 1;
my $LogFileName = $ARGV[0];
my $TimeInt = 60;
my $TemplateFile = "default_html.tpl";
my $WriteMode = ">";
my ($OutDataFile, $OutHTMLFile, $OutPortionsFile, $OutNodesFile);
shift @ARGV;
while (defined ($ARGV[0]) && defined ($ARGV[1]))
{
  $TimeInt = $ARGV[1] if ($ARGV[0] eq "-i");
  $TemplateFile = $ARGV[1] if ($ARGV[0] eq "-t");
  $WriteMode = ">>" if ($ARGV[0] eq "-m" && $ARGV[1] eq "add");
  $OutHTMLFile = $ARGV[1] if ($ARGV[0] eq "-o");
  $OutDataFile = $ARGV[1] if ($ARGV[0] eq "-ot");
  $OutPortionsFile = $ARGV[1] if ($ARGV[0] eq "-op");
  $OutNodesFile = $ARGV[1] if ($ARGV[0] eq "-on");
  shift @ARGV; shift @ARGV;
}

my ($TimeIntStart, $StartTime, $FirstTSRTime, $EndTime, $ServerName, $TaskName, $TaskId, $TaskArgs);

my (%TSRs, %REQs, %ASWs, %TFILEs, %DFILEs, %Repl);
my (%Clusters, %Nodes, %Portions, %Processes, %Sessions, %PortionsStat, %NodesStat, %NodesCount);
my (%SrvTimes, %ClTimes, %MinClTime, %MaxClTime);
my (%PortionsIn, %PortionsOut, %PortionsCount, %MaxPortionIn, %MinPortionIn, %MaxPortionOut, %MinPortionOut);
my (%PortionsOutRealCounter, %PortionsInRealCounter);
my (%MiscDataIn, %MiscDataOut);
my @PortionsInOrderOfASWs;

my %Intervals;
my ($Int, $DemoInt, $PrevDemoInt);
my $TimeIntEnd = 0;

sub getint { return $TimeIntStart + (int(($_[0]-$TimeIntStart)/$TimeInt))*$TimeInt }

print "Parsing $LogFileName\n";
open FH, $LogFileName or die ("$0: can't open $LogFileName!\n");
my $ln=0;
W: while (defined(my $line=<FH>))
{
  chomp $line;
  $ln++;

  my ($DT, $DEST, $REQTYPE, $STR);
  unless (($DT,$DEST,$REQTYPE,$STR) = split (/\t/, $line))
  {
    print STDERR "$0: bad line $ln (general)!\n";
    next;
  }

  my @va = split (/&/, $STR);
  my %VAR=();
  foreach my $v (@va)
  {
    if ($v =~ /([^=]*)=(.*)/)
    { $VAR{$1} = $2; }
    else
    {
      print STDERR "$0: bad line $ln ($REQTYPE)!\n";
      next W;
    }
  }

  $EndTime = $VAR{ETime};

  if ($REQTYPE eq 'START')
  {
    if (defined($VAR{Task}))
    {
      ($StartTime, $ServerName, $TaskName, $TaskId, $TaskArgs) = ($VAR{STime}, $VAR{Server}, $VAR{Task}, '', substr ($VAR{Args}, 1, length($VAR{Args})-2));
    }
    elsif (defined($VAR{TaskName}))
    {
      ($StartTime, $ServerName, $TaskName, $TaskId, $TaskArgs) = ($VAR{STime}, $VAR{Server}, $VAR{TaskName}, $VAR{TaskId}, substr ($VAR{TaskArgs}, 1, length($VAR{TaskArgs})-2));
    }
    $DemoInt = $DT;
    $TimeIntStart = $StartTime;
    next W;
  }

  $Int = getint($VAR{STime});
  $TimeIntEnd = $Int if $TimeIntEnd < $Int;

  if (mkt($DemoInt) + $TimeInt < mkt($DT))
  {
    $DemoInt = $DT;
    # print $DemoInt."\n";
  }

# print STDERR "$Int -> $TimeIntEnd\n";
  if ($REQTYPE eq 'TFILE' || $REQTYPE eq 'DFILE' || $REQTYPE eq 'STAT')
  {
    next W;
  }

  my $Cl;
  if ($VAR{Node} =~ /([^.]+)\.([^.]+).*/)
  { $Cl = $1.'.'.$2 }
  else
  { $Cl = '*' }
  # { $Cl = $VAR{Node} }
  if (!exists ($Clusters{$Cl}))
  {
    $Clusters{$Cl} = 1;
    $SrvTimes{$Cl} = $ClTimes{$Cl} = 0; # $PortionsIn{$Cl} = $PortionsOut{$Cl} = $MinClTime{$Cl} = $MaxClTime{$Cl}
    # $MaxPortionIn{$Cl} = $MinPortionIn{$Cl} = $MaxPortionOut{$Cl} = $MinPortionOut{$Cl} = 0;
  }
  

  ${Sessions{$Cl}}{$VAR{SId}} = 1; # unless (exists (${Sessions{$Cl}}{$VAR{SId}}));
  ${Processes{$Cl}}{$VAR{Node}} = 1; # unless (exists (${Processes{$Cl}}{$VAR{Node}}));
  if ($VAR{Node} =~ /(.*)\.(\d+)/ )
  { ${Nodes{$Cl}}{$1} = 1 } # unless (exists (${Nodes{$Cl}}{$1})) }
  else
  { ${Nodes{$Cl}}{$VAR{Node}} = 1 } # unless (exists (${Nodes{$Cl}}{$VAR{Node}})) }

  $NodesCount{$VAR{Node}} = 1;

  if ($REQTYPE eq 'TSR')
  {
    if (exists(${TSRs{$Cl}}{$Int})) { ${TSRs{$Cl}}{$Int}++ } else { ${TSRs{$Cl}}{$Int} = 1 }
    if (!defined($FirstTSRTime)) { $FirstTSRTime=$VAR{STime} }
    ${NodesStat{$VAR{Node}}}{TSRs} = (exists ${NodesStat{$VAR{Node}}}{TSRs} ? ${NodesStat{$VAR{Node}}}{TSRs} + 1 : 1);
  }

  elsif ($REQTYPE eq 'REQ')
  {
    if (exists(${REQs{$Cl}}{$Int})) { ${REQs{$Cl}}{$Int}++ } else { ${REQs{$Cl}}{$Int} = 1 }
    ${NodesStat{$VAR{Node}}}{REQs} = (exists ${NodesStat{$VAR{Node}}}{REQs} ? ${NodesStat{$VAR{Node}}}{REQs} + 1 : 1);
    my @ps = getpt($VAR{Portion});
    my $psc = @ps;
    next W if $psc == 0;
    my $avg_length = int ($VAR{Length}/$psc);
    my $avg_req_overhead = ($VAR{ETime} - $VAR{STime})/$psc;
    ${PortionsOutRealCounter{$Cl}}{$Int} = (defined ${PortionsOutRealCounter{$Cl}}{$Int} ? ${PortionsOutRealCounter{$Cl}}{$Int} + $psc : $psc);
    # ${PortionsOut{$Cl}}{$Int} = (defined (${PortionsOut{$Cl}}{$Int}) ? ${PortionsOut{$Cl}}{$Int} + $VAR{Length} : $VAR{Length});
    # $MinPortionOut{$Cl} = $VAR{Length} if (!defined ($MinPortionOut{$Cl}) || $MinPortionOut{$Cl} < $VAR{Length});
    # $MaxPortionOut{$Cl} = $VAR{Length} if (!defined ($MaxPortionOut{$Cl}) || $MaxPortionOut{$Cl} > $VAR{Length});
    foreach my $p (@ps)
    {
      # $PortionsCount{$p} = 1 if (!exists($PortionsCount{$p}));
      $PortionsCount{$p} = (exists($PortionsCount{$p}) ? $PortionsCount{$p}+1 : 1);
      ${Portions{$VAR{SId}}}{$p} = $VAR{STime};
      # ${PortionsOut{$Cl}}{$Int} = (defined (${PortionsOut{$Cl}}{$Int}) ? ${PortionsOut{$Cl}}{$Int} + $VAR{Length} : $VAR{Length});
      ${PortionsOut{$Cl}}{$Int} = (defined (${PortionsOut{$Cl}}{$Int}) ? ${PortionsOut{$Cl}}{$Int} + $avg_length : $avg_length);
# print STDERR "$VAR{Length} (Out:) ".${PortionsOut{$Cl}}{$Int}."\n";
      # $MinPortionOut{$Cl} = $VAR{Length} if (!defined ($MinPortionOut{$Cl}) || $MinPortionOut{$Cl} < $VAR{Length});
      $MinPortionOut{$Cl} = $avg_length if (!defined ($MinPortionOut{$Cl}) || $MinPortionOut{$Cl} < $avg_length);
      # $MaxPortionOut{$Cl} = $VAR{Length} if (!defined ($MaxPortionOut{$Cl}) || $MaxPortionOut{$Cl} > $VAR{Length});
      $MaxPortionOut{$Cl} = $avg_length if (!defined ($MaxPortionOut{$Cl}) || $MaxPortionOut{$Cl} > $avg_length);
      # ${PortionsStat{"$p.$VAR{SId}"}}{REQLength} = int ($VAR{Length}/$psc);
      ${PortionsStat{"$p.$VAR{SId}"}}{REQLength} = $avg_length;
      # ${PortionsStat{"$p.$VAR{SId}"}}{REQOverhead} = $VAR{ETime} - $VAR{STime};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQOverhead} = $avg_req_overhead;
      ${PortionsStat{"$p.$VAR{SId}"}}{REQSTime} = $VAR{STime};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQETime} = $VAR{ETime};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQPrepStart} = $VAR{prepStart};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQPrepEnd} = $VAR{prepEnd};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQConnection} = $VAR{prepStart} - $VAR{STime};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQPreparing} = $VAR{prepEnd} - $VAR{prepStart} - $VAR{intCommTime};
      # ${PortionsStat{"$p.$VAR{SId}"}}{REQInternalCommunications} = $VAR{prepEnd} - $VAR{prepStart} - $VAR{APITime};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQInternalCommunications} = $VAR{intCommTime} - $VAR{APITime};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQSending} = $VAR{ETime} - $VAR{prepEnd};
      ${PortionsStat{"$p.$VAR{SId}"}}{REQGeneration} = $VAR{APITime};
      # ${PortionsStat{"$p.$VAR{SId}"}}{REQTime} = mkt ($DT);
    }
  }

  elsif ($REQTYPE eq 'ASW')
  {
    if (exists(${ASWs{$Cl}}{$Int})) { ${ASWs{$Cl}}{$Int}++ } else { ${ASWs{$Cl}}{$Int} = 1 }
    ${NodesStat{$VAR{Node}}}{ASWs} = (exists ${NodesStat{$VAR{Node}}}{ASWs} ? ${NodesStat{$VAR{Node}}}{ASWs} + 1 : 1);
    my @ps = getpt($VAR{Portion});
    my $psc = @ps;
    next W if $psc == 0;
    my $avg_length = int ($VAR{Length}/$psc);
    my $avg_asw_overhead = ($VAR{ETime} - $VAR{STime})/$psc;
    my $avg_client_time = $VAR{Time}/$psc;
    ${PortionsInRealCounter{$Cl}}{$Int} = (defined ${PortionsInRealCounter{$Cl}}{$Int} ? ${PortionsInRealCounter{$Cl}}{$Int} + $psc : $psc);
    # $ClTimes{$Cl} += $VAR{Time};
    # $MinClTime{$Cl} = $VAR{Time} if (!defined($MinClTime{$Cl}) || $VAR{Time} < $MinClTime{$Cl});
    foreach my $p (@ps)
    {
      # if (defined(${Portions{$p}}{S}) && ${Portions{$p}}{S}==1 && ${Portions{$p}}{SId}==$VAR{SId})
      # {
        # ${Portions{$p}}{S} = 2;
        # $SrvTimes{$Cl} += $VAR{ETime} - ${Portions{$p}}{STime}; # !!!
	${PortionsStat{"$p.$VAR{SId}"}}{ServerTime} = $VAR{ETime} - ${Portions{$VAR{SId}}}{$p};
	${PortionsStat{"$p.$VAR{SId}"}}{ClientTime} = $avg_client_time;
#print STDERR "<<".$VAR{Node}.">>\n";
	${PortionsStat{"$p.$VAR{SId}"}}{Node} = $VAR{Node};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWLength} = $avg_length;
	${PortionsStat{"$p.$VAR{SId}"}}{ASWOverhead} = $avg_asw_overhead;
        ${PortionsStat{"$p.$VAR{SId}"}}{ASWSTime} = $VAR{STime};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWETime} = $VAR{ETime};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWPrepStart} = $VAR{prepStart};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWPrepEnd} = $VAR{prepEnd};
        ${PortionsStat{"$p.$VAR{SId}"}}{ASWConnection} = $VAR{prepStart} - $VAR{STime};
        # ${PortionsStat{"$p.$VAR{SId}"}}{ASWInternalCommunications} = $VAR{prepEnd} - $VAR{prepStart} - $VAR{APITime};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWPreparing} = $VAR{prepEnd} - $VAR{prepStart} - $VAR{intCommTime};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWInternalCommunications} = $VAR{intCommTime} - $VAR{APITime};
        ${PortionsStat{"$p.$VAR{SId}"}}{ASWSending} = $VAR{ETime} - $VAR{prepEnd};
	${PortionsStat{"$p.$VAR{SId}"}}{ASWProcessing} = $VAR{APITime};
	${PortionsStat{"$p.$VAR{SId}"}}{ServerOverhead} = ${PortionsStat{"$p.$VAR{SId}"}}{ASWOverhead} + ${PortionsStat{"$p.$VAR{SId}"}}{REQOverhead};
	${PortionsStat{"$p.$VAR{SId}"}}{ClientOverhead} = ${PortionsStat{"$p.$VAR{SId}"}}{ServerTime} - ${PortionsStat{"$p.$VAR{SId}"}}{ClientTime} - ${PortionsStat{"$p.$VAR{SId}"}}{ServerOverhead};
        # ${PortionsStat{"$p.$VAR{SId}"}}{ASWTime} = mkt ($DT);
	$SrvTimes{$Cl} += ${PortionsStat{"$p.$VAR{SId}"}}{ServerTime};
        $ClTimes{$Cl} += $avg_client_time;
	$MinClTime{$Cl} = $avg_client_time if (!defined($MinClTime{$Cl}) || $avg_client_time < $MinClTime{$Cl});
# print STDERR $VAR{Time}."\n";
	$MaxClTime{$Cl} = $avg_client_time if (!defined($MaxClTime{$Cl}) || $avg_client_time > $MaxClTime{$Cl});
# print STDERR $line."\n".$MaxClTime{$Cl}."\n\n";
	${PortionsIn{$Cl}}{$Int} = (defined (${PortionsIn{$Cl}}{$Int}) ? ${PortionsIn{$Cl}}{$Int} + $avg_length : $avg_length);
# print STDERR "$VAR{Length} (In:) ".${PortionsIn{$Cl}}{$Int}."\n";
	$MinPortionIn{$Cl} = $avg_length if (!defined($MinPortionIn{$Cl}) || ($MinPortionIn{$Cl} < $avg_length));
	$MaxPortionIn{$Cl} = $avg_length if (!defined($MaxPortionIn{$Cl}) || ($MaxPortionIn{$Cl} > $avg_length));
      # }
	push (@PortionsInOrderOfASWs, "$p.$VAR{SId}");
    }
  }
# print STDERR getint($VAR{STime})."\n";
}

close FH;

my @ckeys = sort keys %Clusters;
#my $clist = join ("\t", @ckeys);

# print STDERR "\n$TimeIntStart - $TimeIntEnd\n";

$OutDataFile = "$LogFileName.$TaskName.general.txt" if (!defined($OutDataFile));
print "Writing $OutDataFile\n";
open OD, "$WriteMode $OutDataFile" or die ("$0: can't write to $OutDataFile!\n"); 
print OD "T\tTSR,".join("\tTSR,",@ckeys)."\tTSR,SUM\tREQ,".join("\tREQ,",@ckeys)."\tREQ,SUM\tpSent,".join("\tpSent,",@ckeys)."\tpSent,SUM\tASW,".join("\tASW,",@ckeys)."\tASW,SUM\tpReceived".join("\tpReceived,",@ckeys)."\tpReceived,SUM\tALL,".join("\tALL,",@ckeys)."\tALL,SUM\tOUT,".join("\tOUT,",@ckeys)."\tOUT,SUM\tIN,".join("\tIN,",@ckeys)."\tIN,SUM\tINOUT,".join("\tINOUT,",@ckeys)."\tINOUT,SUM\n";
my $i;
$Repl{TSRs} = $Repl{ASWs} = $Repl{REQs} = $Repl{pSent} = $Repl{pReceived} = 0;
my (%TSRsCount, %REQsCount, %ASWsCount, %InData, %OutData, %pREQsCount, %pASWsCount);
foreach my $c (@ckeys) { $TSRsCount{$c} = $REQsCount{$c} = $ASWsCount{$c} = $InData{$c} = $OutData{$c} = $pREQsCount{$c} = $pASWsCount{$c}= 0 }
$Repl{ServerDataOut} = $Repl{ServerDataIn} = 0;
for ($i = $TimeIntStart; $i <= $TimeIntEnd; $i += $TimeInt)
{
# print STDERR $i."\n";
  my ($c, $v);
  my ($c1, $c2, $c3, $c4, $c5, $c6, $c7, $c22, $c33) = (0,0,0,0,0,0,0,0,0);
  my $tm = localtime(int($i));
  #print 1900+$tm->year().'-'.(1+$tm->mon()).'-'.$tm->mday().' '.$tm->hour().':'.$tm->min().':'.$tm->sec()."\t";
  print OD z($tm->mday()).' '.z($tm->hour()).':'.z($tm->min()).':'.z($tm->sec())."\t";
  foreach $c (@ckeys)
    { $v = ifz(${TSRs{$c}}{$i}); print OD $v."\t"; $c1+=$v; $TSRsCount{$c}+=$v; }
  print OD "$c1\t";
  $Repl{TSRs} += $c1;
  foreach $c (@ckeys)
    { $v = ifz(${REQs{$c}}{$i}); print OD $v."\t"; $c2+=$v; $REQsCount{$c}+=$v; }
  print OD "$c2\t";
  $Repl{REQs} += $c2;

  foreach $c (@ckeys)
    { $v = ifz(${PortionsOutRealCounter{$c}}{$i}); print OD $v."\t"; $c22+=$v; $pREQsCount{$c}+=$v; }
  print OD "$c22\t";
  $Repl{pSent} += $c22;

  foreach $c (@ckeys)
    { $v = ifz(${ASWs{$c}}{$i}); print OD $v."\t"; $c3+=$v; $ASWsCount{$c}+=$v; }
  print OD "$c3\t";
  $Repl{ASWs} += $c3;

  foreach $c (@ckeys)
  { $v = ifz(${PortionsInRealCounter{$c}}{$i}); print OD $v."\t"; $c33+=$v; $pASWsCount{$c}+=$v; }
  print OD "$c33\t";
  $Repl{pReceived} += $c33;

  foreach $c (@ckeys)
    { $v = (ifz(${ASWs{$c}}{$i}) + ifz(${REQs{$c}}{$i}) + ifz(${TSRs{$c}}{$i})); print OD $v."\t"; $c4+=$v; }
  print OD "$c4\t";
  foreach $c (@ckeys)
    { $v = ifz(${PortionsOut{$c}}{$i}); print OD $v."\t"; $c5+=$v; $OutData{$c}+=$v; }
  print OD "$c5\t";
  $Repl{ServerDataOut} += $c5;
  foreach $c (@ckeys)
    { $v = ifz(${PortionsIn{$c}}{$i}); print OD $v."\t"; $c6+=$v; $InData{$c}+=$v; }
  print OD "$c6\t";
  $Repl{ServerDataIn} += $c6;
  foreach $c (@ckeys)
    { $v = (ifz(${PortionsIn{$c}}{$i}) + ifz(${PortionsOut{$c}}{$i})); print OD $v."\t"; $c7+=$v; }
  print OD "$c7\n";
}
close OD;

$OutPortionsFile = "$LogFileName.$TaskName.portions.txt" if (!defined($OutPortionsFile));
print "Writing $OutPortionsFile\n";
open OP, "$WriteMode $OutPortionsFile" or die ("$0: can't write to $OutPortionsFile!\n");

print OP "Portion\tNode\tComputingTime\tTotalProcessingTime\tREQLength\tREQConnection\tREQGeneration\tREQInternalCommunications\tREQPreparing\tREQSending\tREQTotalOverhead\tASWLength\tASWConnection\tASWProcessing\tASWInternalCommunications\tASWPreparing\tASWSending\tASWTotalOverhead\tServerTotalOverhead\tClientOverhead\n";

#foreach my $p (sort {$a <=> $b} keys %PortionsCount)
foreach my $p (@PortionsInOrderOfASWs)
{
  my $opl = (defined (${PortionsStat{$p}}{ClientTime}) ? ${PortionsStat{$p}}{ClientTime} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ServerTime}) ? ${PortionsStat{$p}}{ServerTime} : '-')
#  ."\t".${PortionsStat{$p}}{REQSTime}
#  ."\t".${PortionsStat{$p}}{REQPrepStart}
#  ."\t".${PortionsStat{$p}}{REQPrepEnd}
#  ."\t".${PortionsStat{$p}}{REQETime}
  ."\t".${PortionsStat{$p}}{REQLength}
  ."\t".${PortionsStat{$p}}{REQConnection}
  ."\t".${PortionsStat{$p}}{REQGeneration}
  ."\t".${PortionsStat{$p}}{REQInternalCommunications}
  ."\t".${PortionsStat{$p}}{REQPreparing}
  ."\t".${PortionsStat{$p}}{REQSending}
  ."\t".${PortionsStat{$p}}{REQOverhead}
#  ."\t".(defined (${PortionsStat{$p}}{ASWSTime}) ? ${PortionsStat{$p}}{ASWSTime} : '-')
#  ."\t".(defined (${PortionsStat{$p}}{ASWPrepStart}) ? ${PortionsStat{$p}}{ASWPrepStart} : '-')
#  ."\t".(defined (${PortionsStat{$p}}{ASWPrepEnd}) ? ${PortionsStat{$p}}{ASWPrepEnd} : '-')
#  ."\t".(defined (${PortionsStat{$p}}{ASWETime}) ? ${PortionsStat{$p}}{ASWETime} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ASWLength}) ? ${PortionsStat{$p}}{ASWLength} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ASWConnection}) ? ${PortionsStat{$p}}{ASWConnection} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ASWProcessing}) ? ${PortionsStat{$p}}{ASWProcessing} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ASWInternalCommunications}) ? ${PortionsStat{$p}}{ASWInternalCommunications} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ASWPreparing}) ? ${PortionsStat{$p}}{ASWPreparing} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ASWSending}) ? ${PortionsStat{$p}}{ASWSending} : '-')  
  ."\t".(defined (${PortionsStat{$p}}{ASWOverhead}) ? ${PortionsStat{$p}}{ASWOverhead} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ServerOverhead}) ? ${PortionsStat{$p}}{ServerOverhead} : '-')
  ."\t".(defined (${PortionsStat{$p}}{ClientOverhead}) ? ${PortionsStat{$p}}{ClientOverhead} : '-')
  ."\n";
  $opl =~ tr/./,/;
  my ($p2, $sid) = split (/\./, $p);
  $opl = $p2."\t".(defined ${PortionsStat{$p}}{Node} ? ${PortionsStat{$p}}{Node} : '-')."\t".$opl;
  print OP $opl;
}
close OP;

$OutNodesFile = "$LogFileName.$TaskName.nodes.txt" if (!defined($OutNodesFile));
print "Writing $OutNodesFile\n";
open ON, "$WriteMode $OutNodesFile" or die ("$0: can't write to $OutNodesFile!\n");
print ON "Node\tTSRs\tREQs\tASWs\n";
foreach my $n (sort keys %NodesCount)
{ print ON $n."\t".ifz(${NodesStat{$n}}{TSRs})."\t".ifz(${NodesStat{$n}}{REQs})."\t".ifz(${NodesStat{$n}}{ASWs})."\n" }
close ON;

$Repl{LogFileName} = $LogFileName;
$Repl{TaskName} = $TaskName;
$Repl{TaskArgs} = $TaskArgs;
$Repl{ServerName} = $ServerName;
$Repl{StartTime} = lct($StartTime);
$Repl{FirstTSRTime} = lct($FirstTSRTime);
$Repl{EndTime} = lct($EndTime);
$Repl{TotalTime} = mt($EndTime-$StartTime);
$Repl{TotalWorkingTime} = mt($EndTime-$FirstTSRTime);
$Repl{OutDataFile} = $OutDataFile;
my ($ClTimesSum, $SrvTimesSum) = (0, 0);
$Repl{Nodes} = $Repl{Sessions} = $Repl{Processes} = 0;
foreach my $c (keys %Clusters)
{
  $Repl{Nodes} += scalar keys %{$Nodes{$c}};
  $Repl{Sessions} += scalar keys %{$Sessions{$c}};
  $Repl{Processes} += scalar keys %{$Processes{$c}};
  $SrvTimesSum += $SrvTimes{$c};
  $ClTimesSum += $ClTimes{$c};
}
$Repl{ServerTime} = mt($SrvTimesSum);
$Repl{ClientsTime} = mt($ClTimesSum);
$Repl{ServerEfficiency} = sprintf ("%.2f", 100*$ClTimesSum/$SrvTimesSum);
$Repl{Portions} = scalar keys %PortionsCount;
#$Repl{ClientsEfficiency} = sprintf ("%.2f", 100*$Repl{ASWs}/$Repl{REQs});
$Repl{ClientsEfficiency} = sprintf ("%.2f", 100*$Repl{pReceived}/$Repl{pSent});
$Repl{MinTaskTime} = mt(min(values %MinClTime));
$Repl{MaxTaskTime} = mt(max(values %MaxClTime));
#$Repl{AverageTaskTime} = mt($ClTimesSum/$Repl{ASWs});
$Repl{AverageTaskTime} = mt($ClTimesSum/$Repl{pReceived});
$Repl{ServerDataSum} = $Repl{ServerDataIn} + $Repl{ServerDataOut};
$Repl{ServerDataIn} = sprintf ("%.3f", $Repl{ServerDataIn}/(1024*1024));
$Repl{ServerDataOut} = sprintf ("%.3f", $Repl{ServerDataOut}/(1024*1024));
$Repl{ServerDataSum} = sprintf ("%.3f", $Repl{ServerDataSum}/(1024*1024));

my ($tr1, $td1, $td2, $tr2, $tb1, $tb2) = qw (<tr> <td> </td> </tr> <tbody> </tbody>);
my ($th1, $th2) = ("<tr class=th>", "</tr>");
my @Ithead = ("Cluster", "Nodes", "Processes", "Sessions", "Clients time",
"Server time", "Server eff.", "Code requests", "Sent portions", 
"Received portions", "Lost portions", "Participation", "Clients eff.", "MIN prt. time", "MAX prt. time", "AVG prt. time");
my @Dthead = ("Cluster", "Output", "Input", "Summary", "MIN task in", 
"MAX task in", "AVG task in", "MIN task out", "MAX task out", "AVG task out");
$Repl{ClustersInfoTableHead} = $th1.$td1.join($td2.$td1,@Ithead).$td2.$th2;
$Repl{ClustersDataTableHead} = $th1.$td1.join($td2.$td1,@Dthead).$td2.$th2;
$Repl{ClustersInfoTable} = $Repl{ClustersDataTable} = '';
for my $c (keys %Clusters)
{
  $Repl{ClustersInfoTable} .= $tb1.$tr1.$td1.$c.$td2.
  $td1.(scalar keys %{$Nodes{$c}}).$td2.
  $td1.(scalar keys %{$Sessions{$c}}).$td2.
  $td1.(scalar keys %{$Processes{$c}}).$td2.
  $td1.mt($ClTimes{$c}).$td2.
  $td1.mt($SrvTimes{$c}).$td2.
  $td1.($SrvTimes{$c}!=0 ? sprintf("%.2f", 100*$ClTimes{$c}/$SrvTimes{$c}).'%' : '-').$td2.
  $td1.$TSRsCount{$c}.$td2.
#  $td1.$REQsCount{$c}.$td2.
  $td1.$pREQsCount{$c}.$td2.
#  $td1.$ASWsCount{$c}.$td2.
  $td1.$pASWsCount{$c}.$td2.
#  $td1.($REQsCount{$c}-$ASWsCount{$c}).' ('.sprintf("%.2f",($REQsCount{$c}-$ASWsCount{$c})/$REQsCount{$c}*100).'%)'.$td2.
  $td1.($pREQsCount{$c}!=0 ? ($pREQsCount{$c}-$pASWsCount{$c}).' ('.sprintf("%.2f",($pREQsCount{$c}-$pASWsCount{$c})/$pREQsCount{$c}*100).'%)' : '-').$td2.
#  $td1.sprintf("%.2f", 100*$ASWsCount{$c}/$Repl{ASWs}).'%'.$td2.
  $td1.($Repl{pReceived}!=0 ? sprintf("%.2f", 100*$pASWsCount{$c}/$Repl{pReceived}).'%' : '-').$td2.
#  $td1.sprintf("%.2f", 100*$ASWsCount{$c}/$REQsCount{$c}).'%'.$td2.
  $td1.($pREQsCount{$c}!=0 ? sprintf("%.2f", 100*$pASWsCount{$c}/$pREQsCount{$c}).'%' : '-').$td2.
  $td1.mt($MinClTime{$c}).$td2.
  $td1.mt($MaxClTime{$c}).$td2.
#  $td1.mt($ClTimes{$c}/$ASWsCount{$c}).$td2.$tr2;
  $td1.($pASWsCount{$c}!=0 ? mt($ClTimes{$c}/$pASWsCount{$c}) : '-').$td2.$tr2;
  $Repl{ClustersDataTable} .= $tb1.$tr1.$td1.$c.$td2.
  $td1.sprintf("%.3f Mb",$OutData{$c}/(1024*1024)).$td2.
  $td1.sprintf("%.3f Mb",$InData{$c}/(1024*1024)).$td2.
  $td1.sprintf("%.3f Mb",($OutData{$c}+$InData{$c})/(1024*1024)).$td2.
  $td1.sprintf("%.3f Mb",$MinPortionOut{$c}/(1024*1024)).$td2.
  $td1.sprintf("%.3f Mb",$MaxPortionOut{$c}/(1024*1024)).$td2.
  $td1.($REQsCount{$c}!=0 ? sprintf("%.3f Mb",$OutData{$c}/$REQsCount{$c}/(1024*1024)) : '-').$td2.
  $td1.sprintf("%.3f Mb",$MinPortionIn{$c}/(1024*1024)).$td2.
  $td1.sprintf("%.3f Mb",$MaxPortionIn{$c}/(1024*1024)).$td2.
  $td1.($ASWsCount{$c}!=0 ? sprintf("%.3f Mb",$InData{$c}/$ASWsCount{$c}/(1024*1024)) : '-').$td2.$tr2;
}
$Repl{ClustersInfoTable} .= $tb2;
$Repl{ClustersDataTable} .= $tb2;

$OutHTMLFile = "$LogFileName.$TaskName.html" if (!defined($OutHTMLFile));
open TF, $TemplateFile or die ("$0: can't read template from $TemplateFile!\n");
print "Writing $OutHTMLFile using template $TemplateFile\n";
open OUT, "$WriteMode $OutHTMLFile" or die ("$0: can't write to $OutHTMLFile!\n");
while (defined(my $line=<TF>))
{
  foreach my $k (keys %Repl) { $line =~ s/<%$k%>/$Repl{$k}/g }
  print OUT $line;
}
close OUT;
close TF;
print "Done.\n";
#print STDERR mt (mkt($EndTime)-mkt($StartTime));
