#!/usr/bin/perl -W

use strict;
no strict 'subs';
use warnings;

#use Carp;
#$SIG{__WARN__} = \&confess;

BEGIN {
  if ($^O eq 'MSWin32')
  {
    eval "require Win32::Process" or die ("XQServ: can't load Win32::Process!\n"); 
    print "XQServ: MSWin32 modules loaded\n";
  }
  else
  {
    eval "require IO::Socket::UNIX" or die ("XQServ: can't load IO::Socket::UNIX!\n");
    eval "require IO::Handle::Record" or die ("XQServ: can't load IO::Handle::Record!\n");
    print "XQServ: UNIX modules loaded\n";
  }
}

use IO::Select;
use HTTP::Daemon;
use XML::Parser;
use File::Copy;
use POSIX ":sys_wait_h";
use FindBin qw($Bin);
use lib "$Bin";
use sigtrap qw(die normal-signals error-signals);
# use Data::Dumper;
use XQUtils;

my %RunningTasksPIDs = ();
my %RunningTasksClients = ();
my %RunningTasksPorts = ();
my $RunningTasks = 0;
my ($nextTaskForClient, $lastRedirectedTask, $nextPort) = (0, 0, 0);
my ($c_pid, $c_uid, $c_gid);
my $PID = $$;

our %IniSettings;
our @Queue;

$SIG{CHLD} = 'IGNORE';
$SIG{PIPE} = 'IGNORE';
# $SIG{TERM} = $SIG{INT} = $SIG{HUP} = \&onExit;

my $iniFile = (defined($ARGV[0]) ? $ARGV[0] : $Bin.'/XQServ.ini');
die ("XQServ: Can't read $iniFile!\n") unless (-f $iniFile);
XQUtils::parseIniFile ($iniFile, \%IniSettings);

my $xIniFilePath = (defined $IniSettings{XQServDataPath} ? $IniSettings{XQServDataPath} : $Bin).'/xqserv_tasks/';
unless (-d $xIniFilePath)
{
  mkdir $xIniFilePath or die ("XQServ: can't create $xIniFilePath!\n");
  print "XQServ: created $xIniFilePath\n";
}
die ("XQServ: can't write to $xIniFilePath") unless (-w $xIniFilePath);

my $counterFile = (defined $IniSettings{XQServDataPath} ? $IniSettings{XQServDataPath} : $Bin).'/xqserv_counter.dat';
my $Counter = XQUtils::getCounter($counterFile);

$IniSettings{MaxRunningTasks} = 1 if !defined $IniSettings{MaxRunningTasks};
$IniSettings{ClientsRedirect} = 'Equal' if !defined $IniSettings{ClientsRedirect};

my $HTTPInterface = HTTP::Daemon->new(LocalHost => $IniSettings{XQServHost}, LocalPort => $IniSettings{XQServPort}, Type => SOCK_STREAM, ReuseAddr => 1, Listen => SOMAXCONN, Proto => "tcp", Blocking => 0)
  or die ("XQServ: Can't create HTTP daemon for control server!\n");
print "XQServ: monitoring at http://$IniSettings{XQServHost}:$IniSettings{XQServPort}\n";

my $ControlServer;
if (defined($IniSettings{XQServSocket}))
{
  if (-S $IniSettings{XQServSocket})
  {
    unlink $IniSettings{XQServSocket};
    print "XQServ: cleared old socket $IniSettings{XQServSocket}\n";
  }
  my $perm = (defined($IniSettings{XQServSocketPerm}) ? oct($IniSettings{XQServSocketPerm}) : 0770);
  my $ou = umask;
# printf STDERR "%lo\n%lo\n", $ou, $perm;
  umask(0777 - $perm);
  $ControlServer = IO::Socket::UNIX->new(Type => SOCK_STREAM, Local => $IniSettings{XQServSocket})
    or die ("XQServ: Can't create $IniSettings{XQServSocket} for control server!\n");
  if (defined($IniSettings{XQServSocketGID})) { chown (-1, $IniSettings{XQServSocketGID}, $IniSettings{XQServSocket}) }
  $ControlServer->listen
    or die ("XQServ: Can't listen on $IniSettings{XQServSocket}!\n");
  umask($ou);
}
else
{
  $ControlServer = $HTTPInterface;
}

my $CliManager = HTTP::Daemon->new(LocalHost => $IniSettings{CliManagerHost}, LocalPort => $IniSettings{CliManagerPort}, Type =>SOCK_STREAM, ReuseAddr => 1, Listen => SOMAXCONN, Proto => "tcp", Blocking => 0)
  or die ("XQServ: Can't create incoming socket for client manager!\n");
print "XQServ: clients manager at http://$IniSettings{CliManagerHost}:$IniSettings{CliManagerPort}\n";

my $CS = new IO::Select ($ControlServer);
my $CM = new IO::Select ($CliManager);
my $WI = ($ControlServer!=$HTTPInterface ? new IO::Select ($HTTPInterface) : undef);

my @GCs = ();
my @XQCs = ();
my @WIs = ();
my $xqc;
my $cli;
my $wb;
my @CliKids;

my ($XComMinPort, $XComCurrentPort);
$XComMinPort = $XComCurrentPort = $IniSettings{XComMinPort};

my @FreePorts = ();
for (my $t=0; $t<$IniSettings{MaxRunningTasks}; $t++)
{ push (@FreePorts, $IniSettings{XComMinPort}+2*$t) }

XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $ControlServer->sockhost())."\t[SRV]\tSTART");

while (1)
{
  if (@XQCs = $CS->can_read(0))
  {
    foreach my $fh1 (@XQCs)
    {
      if ($fh1 == $ControlServer)
      {
	$xqc = $ControlServer->accept;
	$CS->add($xqc);
      }
      else
      {
	my ($xqcRequest, $decodedContent);
	if (defined($IniSettings{XQServSocket}))
	{
	  ($c_pid, $c_uid, $c_gid) = $xqc->peercred();
	  while (my $line = $xqc->getline) 
	  { 
	    chomp $line;
	    last if $line eq '';
# print STDERR "|$line|"; 
	    $decodedContent .= $line; 
	  }
	}
	else
	{
    	  $xqcRequest = $xqc->get_request();
	  $decodedContent = $xqcRequest->decoded_content();
	}
#print STDERR $xqcRequest->decoded_content();
    	my $p = new XML::Parser(Style => 'Tree', Pkg => 'XQServ');
    	my $xqcRequestParsed = substr($decodedContent,0,1) eq '<' ? $p->parse($decodedContent) : undef;
    	my $resp = HTTP::Response->new();
#print STDERR Dumper $xqcRequestParsed;
	if (!defined ($xqcRequestParsed))
	{
      	  my %Form = $xqcRequest->url->query_form;
      	  $resp->code(200);
      	  $resp->content(XQUtils::generateHTML(defined $Form{r} ? $Form{r} : 0));
      	  XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $xqc->peerhost())."\t[Web]\tHTTP show");
	}
    	elsif ($xqcRequestParsed->[0] eq 'show') 
	{
	  my $user = (defined($c_uid) ? getpwuid($c_uid) : $xqcRequestParsed->[1]->[0]->{user});
      	  my $q='';
      	  foreach my $k (@Queue)
      	  {
# print STDERR "${$k}{ID}\t${$k}{STATE}\n";
	    $q.="<task id=\"${$k}{ID}\" name=\"${$k}{DESC}{taskName}\" state=\"${$k}{STATE}\" user=\"${$k}{USER}\" qtime=\"${$k}{QTIME}\" stime=\"${$k}{STIME}\" sshost=\"".(${$k}{STATE} eq 'R' ? $IniSettings{XComSsHost} : '')."\" ssport=\"${$k}{SSPORT}\" />";
      	  }
      	  $resp->code(203);
      	  $resp->content("<tasks>$q</tasks>");
      	  XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $xqc->peerhost())."\t[XQC]\t$user show");
    	}
	elsif ($xqcRequestParsed->[0] eq 'add')
	{
      	  my $taskDesc = $xqcRequestParsed->[1]->[2]->[0];
	  if (defined(${$taskDesc}{xFile}))
	  {
	    ${$taskDesc}{taskArgs} = (defined(${$taskDesc}{taskArgs}) ? ${$taskDesc}{xFile}.' '.${$taskDesc}{taskArgs} : ${$taskDesc}{xFile});
	  }
	  my $user = (defined($c_uid) ? getpwuid($c_uid) : $xqcRequestParsed->[1]->[0]->{user});
      	  my %Task;
#print STDERR Dumper %Task;
      	  $Task{DESC} = \%{$taskDesc};
      	  $Task{USER} = $user;
      	  $Task{ID} = $Counter = XQUtils::updateCounter($Counter, $counterFile);
      	  $Task{QTIME} = time();
      	  $Task{STIME} = 0;
      	  $Task{STATE} = 'Q';
      	  $Task{PID} = 0;
	  $Task{SSPORT} = 0;
	  # $Task{CLIENTS} = 0;
      	  push (@Queue, \%Task);
# print STDERR Dumper @Queue;
      	  XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $ControlServer->sockhost())."\t[SRV]\tADD $Task{ID}");
      	  $resp->code(200);
      	  #$resp->content("XQServ: Task was queued with ID=$Task{ID}.");
      	  $resp->content("<task id=\"$Task{ID}\" />");
      	  XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $xqc->peerhost())."\t[XQC]\t$user add");
    	}
    	elsif ($xqcRequestParsed->[0] eq 'kill')
    	{
      	  my $id = $xqcRequestParsed->[1]->[0]->{id};
      	  my $user = (defined($c_uid) ? getpwuid($c_uid) : $xqcRequestParsed->[1]->[0]->{user});
      	  my $i;
      	  for ($i=0; $i<@Queue; $i++)
      	  {
            last if (${$Queue[$i]}{ID} == $id);
      	  }
      	  if ($i < @Queue)
      	  {
	    if (${$Queue[$i]}{USER} eq $user)
	    {
	      if (${$Queue[$i]}{STATE} eq 'R')
	      {
		killPID(${$Queue[$i]}{PID});
	    	# $resp->content("XQServ: Task $id was stopped and removed from queue.");
	    	$resp->code(201);
	    	$resp->content('<ok />');
	      }
	      else
	      {
	        splice (@Queue, $i, 1);
	      	$resp->code(202);
	      	# $resp->content("XQServ: Task $id removed from queue.");
	      	$resp->content('<ok />');
	      }
	    }
	    else
	    {
	      $resp->code(401);
	      # $resp->content("XQServ: Task $id was not removed.");
	      $resp->content('<err />');
	    }
          }
      	  else
      	  { 
            $resp->code(404);
	    #$resp->content("XQServ: Can't find task with ID=$id.");
	    $resp->content('<err />');
      	  }
      	  XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $xqc->peerhost())."\t[XQC]\t$user kill $id");
        }
      	else
      	{
      	  $resp->code(300);
      	}
	if (defined($IniSettings{XQServSocket}))
	{
	  $xqc->print($resp->code."\n".$resp->content);
	}
	else
	{
      	  $xqc->send_response($resp);
	}
      	$CS->remove($fh1);
      	$fh1->close();
      }
    }
  }
  else 
  {
#print STDERR "!";
    if (($RunningTasks < $IniSettings{MaxRunningTasks}) && (@Queue!=0))
    {
      my $nextTask;
      for ($nextTask=0; $nextTask<@Queue; $nextTask++)
      { last if (${$Queue[$nextTask]}{STATE} eq 'Q') }
      if ($nextTask < @Queue)
      {
        # my ($ssHost, $prHost) = ($IniSettings{XComSsHost}, $IniSettings{XComPrHost});
	my $prHost = $IniSettings{XComPrHost} if defined($IniSettings{XComPrHost});
	my $ssHost = $IniSettings{XComSsHost};
        my $ssPort = shift (@FreePorts);
        my $prPort = $ssPort + 1;
	my $prSock = $IniSettings{XComPrSocketPath}.'/xcom.sock.'.$prPort if defined($IniSettings{XComPrSocketPath}); 
#print STDERR "ssPort: $ssPort, prPort: $prPort\n";
        # my $xIniFile = $IniSettings{xFilesPath}.'/'.${$Queue[$nextTask]}{DESC}{taskName}.'.'.${$Queue[$nextTask]}{ID}.'.ini';
        my $xIniFile = $xIniFilePath.${$Queue[$nextTask]}{DESC}{taskName}.'.'.${$Queue[$nextTask]}{ID}.'.ini';
        open F, "> $xIniFile" or die ("XQServ: Can't write to $xIniFile\n");
        # foreach my $k (qw(prHost prPort ssHost ssPort)) { print F "$k = $IniSettings{$k}\n" }
        print F "ssHost = $ssHost\nssPort = $ssPort\n"
          .(defined($prHost) ? "prHost = $prHost\nprPort = $prPort\n" : "prSock = $prSock\n")
          ."taskId = ${$Queue[$nextTask]}{ID}\n"
          .(defined($IniSettings{XComSinglePortion}) ? "singlePortion = $IniSettings{XComSinglePortion}\n" : '')
          .(defined($IniSettings{XComPortionWait}) ? "portionWait = $IniSettings{XComPortionWait}\n" : '')
          .(defined($IniSettings{XComWriteUser}) && ($IniSettings{XComWriteUser} == 1) ? "xUser = ${$Queue[$nextTask]}{USER}\n" : '');
        while (my ($k,$v)=each(%{${$Queue[$nextTask]}{DESC}})) { print F "$k = $v\n" if (($k ne 'prHost') && ($k ne 'prPort') && ($k ne 'prSock') && ($k ne 'ssHost') && ($k ne 'ssPort')) } 
        close F;
        my $pid_x;
        if ($^O ne 'MSWin32')
        {
          $pid_x = fork();
          die ("XQServ: can't fork!\n") if !defined $pid_x;
          if ($pid_x == 0)
          {
	    setpgrp ($$, $$);
	    # my $run = $IniSettings{startXCom}.' '.$xIniFile;
	    my $run = $Bin.'/start_xcom.sh '.$IniSettings{XComServerPath}.' '.$IniSettings{XComServerName}.' '.$xIniFile;
            exec ($run) or die ("XQServ: Can't execute $run\n");
	    exit();
          }
        }
        else
        {
          my $ProcessObj;
          # Win32::Process::Create($ProcessObj, "C:\\strawberry\\perl\\bin\\perl.exe", "perl D:\\dev-xcom\\gserv\\XServ.pl $xIniFile", 0, NORMAL_PRIORITY_CLASS, "D:\\dev-xcom\\gserv\\") or die ("");
	  Win32::Process::Create($ProcessObj, $^X, 'perl '.$IniSettings{XComServerPath}.'\\'.$IniSettings{XComServerName}.' '.$xIniFile, 0, NORMAL_PRIORITY_CLASS | CREATE_NO_WINDOW, $IniSettings{XComServerPath}) 
	    or die ("XQServ: Can't create process!\n");
          $pid_x = $ProcessObj->GetProcessID();
        }
        ${$Queue[$nextTask]}{STATE} = 'R';
        ${$Queue[$nextTask]}{PID} = $RunningTasksPIDs{${$Queue[$nextTask]}{ID}} = $pid_x;
        ${$Queue[$nextTask]}{STIME} = time();
        ${$Queue[$nextTask]}{SSPORT} = $RunningTasksPorts{${$Queue[$nextTask]}{ID}} = $ssPort;
        $RunningTasksClients{${$Queue[$nextTask]}{ID}} = 0;
        $RunningTasks++;
        # $RuntimeTaskCounter++;
        XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $ControlServer->sockhost())."\t[SRV]\tRUN ${$Queue[$nextTask]}{ID}");
      }
    }
#print STDERR "[ @{[%RunningTasksPIDs]} ]\n";
    foreach my $p (keys %RunningTasksPIDs)
    {
#print STDERR "$p >> $RunningTasksPIDs{$p}\n";
     if (waitpid($RunningTasksPIDs{$p}, WNOHANG))
      {
        # shift (@Queue);
      	my $i;
	for ($i=0; $i<@Queue; $i++)
      	{ last if (${$Queue[$i]}{ID} == $p) }
	push (@FreePorts, ${$Queue[$i]}{SSPORT});
	splice (@Queue, $i, 1);
        delete ($RunningTasksPIDs{$p});
	delete ($RunningTasksClients{$p});
	delete ($RunningTasksPorts{$p});
	$RunningTasks--;
        XQUtils::logPrint ((defined($IniSettings{XQServSocket}) ? $IniSettings{XQServSocket} : $ControlServer->sockhost())."\t[SRV]\tEND $p");
	# next;
#print STDERR "Queue = ".@Queue." RunningTasks = $RunningTasks\n";
      }
    }
  }

  if (@GCs = $CM->can_read(0))
  {
    foreach my $fh2 (@GCs)
    {
      if ($fh2 == $CliManager)
      {
	$cli = $CliManager->accept;
	$CM->add($cli);
      }
      else
      {
	($nextTaskForClient, $nextPort) = (0, 0);
  	if ($RunningTasks > 0)
  	{
	  if ($IniSettings{MaxRunningTasks} == 1)
	  {
# print STDERR Dumper %RunningTasksPorts;
	    # ($nextTaskForClient, $nextPort) = each (%RunningTasksPorts);
	    my @TP = %RunningTasksPorts;
	    ($nextTaskForClient, $nextPort) = ($TP[0], $TP[1]);
	  }
	  elsif ($IniSettings{ClientsRedirect} eq 'Equal')
	  {
	    my @O = sort { $RunningTasksClients{$a} > $RunningTasksClients{$b} } keys %RunningTasksClients;
#foreach my $a (@O) {print STDERR "$a >> ".$RunningTasksClients{$a}."\n"; }
    	    $nextTaskForClient = $O[0];
    	    $RunningTasksClients{$nextTaskForClient}++;
    	    $nextPort = $RunningTasksPorts{$nextTaskForClient};
    	    # $lastRedirectedTask = $nextTaskForClient;
	  }
  	}
	my $cliRequest = $cli->get_request();
	if (!defined $cliRequest) {$CM->remove($fh2); $fh2->close; next }
	XQUtils::logPrint ($cli->peerhost()."\t[CL]\tTSR".($nextTaskForClient != 0 ? " >> $nextTaskForClient" : ''));
    	my $pid_c = fork();
    	die ("XQServ: can't fork (CM)!\n") if !defined $pid_c;
    	if ($pid_c == 0)
    	{
      	  # $CliManager->close(); # if ($^O eq 'MSWin32');
      	  # my $cliRequest = $cli->get_request();
      	  my $resp = HTTP::Response->new();
	  if ( ($RunningTasks > 0) && ($IniSettings{ClientsRedirect} eq 'Equal') && ($nextPort != 0) && ($nextTaskForClient != 0) )
      	  {
	    $resp->code(301);
	    $resp->header('Location' => "http://$IniSettings{XComSsHost}:$nextPort");
            $resp->content("<task-redirect order=\"fallback\"><url>http://$IniSettings{XComSsHost}:$nextPort</url></task-redirect>");
      	  }
	  elsif ( ($RunningTasks > 0) && ($IniSettings{ClientsRedirect} eq 'Conditions') )
	  {
	    my $dcC = $cliRequest->decoded_content();
	    my $c = new XML::Parser(Style => 'Tree', Pkg => 'XQServ');
	    my $cliRequestParsed = $c->parse($dcC);
	    $nextTaskForClient = XQUtils::testConditions (\@Queue, $cliRequestParsed);
	    if ($nextTaskForClient == 0)
	    {
	      $resp->code(200);
	      $resp->content("<initialize />");
	    }
	    else
	    {
	      $nextPort = $RunningTasksPorts{$nextTaskForClient};
	      $resp->code(301);
	      $resp->header('Location' => "http://$IniSettings{XComSsHost}:$nextPort");
	      $resp->content("<task-redirect order=\"fallback\"><url>http://$IniSettings{XComSsHost}:$nextPort</url></task-redirect>");
	    }
	  }
	  else
	  {
	    $resp->code(200);
	    $resp->content("<initialize />");
	  }

      	  $cli->send_response($resp);
      	  exit();
    	}
    	else
    	{
          # $cli->close();
	  if ($^O eq 'MSWin32')
	  {
    	    push (@CliKids, $pid_c);
    	    my @kids;
    	    foreach my $childpid (@CliKids) 
    	    {
      	      if (waitpid ($childpid,WNOHANG)!=$childpid) 
      	      {
                push(@kids,$childpid);
      	      }
    	    }
    	    @CliKids=@kids;
	  }
    	}
      	$CM->remove($fh2);
      	$fh2->close();
      }
    }
  }
  
  if (($ControlServer!=$HTTPInterface) && (@WIs = $WI->can_read(0)))
  {
    foreach my $fh3 (@WIs)
    {
      if ($fh3 == $HTTPInterface)
      {
        $wb = $HTTPInterface->accept;
	$WI->add($wb);
      }
      else
      {
	my $wbRequest = $wb->get_request();
	my $resp = HTTP::Response->new();
	my %Form = $wbRequest->url->query_form;
        $resp->code(200);
	$resp->content(XQUtils::generateHTML(defined $Form{r} ? $Form{r} : 0));
	$wb->send_response($resp);
	XQUtils::logPrint ($wb->peerhost()."\t[Web]\tHTTP show");
	$WI->remove($fh3);
	$fh3->close();
      }
    }
  }
}

sub onExit
{
  if ((defined($IniSettings{XQServSocket})) && ( -S $IniSettings{XQServSocket}))
  { 
    unlink $IniSettings{XQServSocket};
    print "XQServ: cleared $IniSettings{XQServSocket}\n";
  }
  if (defined($IniSettings{KillXComOnExit}) && ($IniSettings{KillXComOnExit} == 1))
  {
    foreach my $q (@Queue)
    {
      if ($q->{STATE} eq 'R')
      {
	killPID($q->{PID});
        print "XQServ: killed X-Com server (PID ".$q->{PID}.") at http://$IniSettings{XComSsHost}:".$q->{SSPORT}."\n";
      }
    }
  }
}

sub killPID
{
  my $pid = $_[0];
  if ($^O ne 'MSWin32')
  { kill -9, getpgrp ($pid) }
  else
  {
    # kill 9, ${$Queue[$i]}{PID}.
    Win32::Process::KillProcess($pid, 0);
  }
}

END {
  onExit() if ($$ == $PID);
}