package XQUtils;
use POSIX;
# use Data::Dumper;
use strict;
# no strict 'refs';

sub parseIniFile
{
  my ($f, $s) = @_;
  open (F, "< $f");
  while (<F>)
  {
    chomp;
    s/#.*//;
    s/^\s+//;
    s/\s+$//;
    next unless length;
    my ($var, $value) = split(/\s*=\s*/, $_, 2);
    ${$s}{$var} = $value;
  }
  close (F);
}

sub getCounter
{
  my $cf = $_[0];
  my $c;
  if (! -f $cf)
  {
    $c = 1;
    open F, "> $cf";
    print F $c;
  }
  else
  {
    open F, "< $cf";
    $c = int (<F>);
  }
  close F;
  return ($c);
}

sub updateCounter
{
  my ($c, $cf) = @_;
  $c++;
  open F, "> $cf";
  print F $c;
  close F;
  return $c;
}

sub logPrint
{
  print STDERR strftime("%Y-%m-%d %H:%M:%S",localtime)."\t".$_[0]."\n";
}

sub generateHTML
{
  my ($refresh) = @_;
  my %st = (R => 'Running', Q => 'Waiting');
  my %c = (R => 0, Q => 0);
  my $out = "<html>
<head>
".($refresh !=0 ? "<meta http-equiv=refresh content=$refresh>\n" : '')."<title>XQServ@".$::IniSettings{XQServHost}."</title>
<style>
tr { font-size: 12px; font-family: Arial }
tr.header { background-color: #000000; color: #FFFFFF }
tr.data1 { background-color: #CCCCCC; color: #000000 }
</style>
</head>
<body>
<table width=100% border=1>
<tr class=header><td>Task ID</td><td>Task name</td><td>State</td><td>User</td><td>Queued</td><td>Started</td><td>URL</td></tr>
";
foreach my $k (@::Queue)
{
  $out.="<tr><td>${$k}{ID}</td><td>${$k}{DESC}{taskName}</td><td>$st{${$k}{STATE}}</td><td>${$k}{USER}</td><td>".strftime("%Y-%m-%d %H:%M:%S",localtime(${$k}{QTIME}))."</td><td>".(${$k}{STIME} == 0 ? '&nbsp;' : strftime("%Y-%m-%d %H:%M:%S",localtime(${$k}{STIME})))."</td><td>".(${$k}{STATE} eq 'R' ? "<a target=_blank href=http://$::IniSettings{XComSsHost}:${$k}{SSPORT}>http://$::IniSettings{XComSsHost}:${$k}{SSPORT}</a>" : '&nbsp;')."</td></tr>";
  $c{${$k}{STATE}}++; 
}
$out .= "
</table>
<br><div align=center>
<table border=1>
<tr class=header><td colspan=2>General info</td></tr>
<tr class=data1><td align=right><b>Tasks: </b></td><td>".($c{R}+$c{Q})." total, $c{R} running, $c{Q} waiting</td></tr>
<tr class=data1><td align=right><b>Simultaneous tasks: </b></td><td>$::IniSettings{MaxRunningTasks}</td></tr>
<tr class=data1><td align=right><b>Clients manager at: </b></td><td><a href=http://$::IniSettings{CliManagerHost}:$::IniSettings{CliManagerPort} onclick='return false'>http://$::IniSettings{CliManagerHost}:$::IniSettings{CliManagerPort}</a></td></tr>
</table>
</div>
</body>
</html>";
return $out;
}

sub testConditions
{
  my ($Q, $r) = @_;
  my $nodeId = $r->[1]->[2]->[0]->{id};
  my $cpuPerf = $r->[1]->[2]->[2]->[0]->{perf};
  my $cpuNumCores = $r->[1]->[2]->[2]->[0]->{'num-cores'};
  my $cpuFreq = $r->[1]->[2]->[2]->[0]->{freq};
  my $cpuType = $r->[1]->[2]->[2]->[0]->{type};
  my $cpuVendor = $r->[1]->[2]->[2]->[0]->{vendor};
  my $ramSize = $r->[1]->[2]->[4]->[0]->{size};
  my $osType = $r->[1]->[2]->[6]->[0]->{type};
# print STDERR "$nodeId\n$cpuPerf\n$cpuNumCores\n$cpuFreq\n$cpuType\n$cpuVendor\n$ramSize\n$osType\n";
  foreach my $q (@$Q)
  {
    last if $q->{STATE} ne 'R';
    my $d = $q->{DESC};
    my ($k, $v);
    my ($cc, $cct) = (0, 0);
    while (($k,$v) = each (%$d))
    {
      next if !($k =~ /^node/);
      $cct++ if (($k eq 'nodeMinCPUFreq') && ($cpuFreq >= $v))
        || (($k eq 'nodeMaxCPUFreq') && ($cpuFreq <= $v))
        || (($k eq 'nodeMinCPUPerf') && ($cpuPerf >= $v))
        || (($k eq 'nodeMaxCPUPerf') && ($cpuPerf <= $v))
        || (($k eq 'nodeMinRAMSize') && ($ramSize >= $v))
        || (($k eq 'nodeMaxRAMSize') && ($ramSize <= $v))
        || (($k eq 'nodeMinCPUCores') && ($cpuNumCores >= $v))
        || (($k eq 'nodeMaxCPUCores') && ($cpuNumCores <= $v));
      if ($k eq 'nodeCluster')
      {
        my @clu = split (/ /, $v);
        foreach my $cl (@clu)
        {
          if ($nodeId =~ /$cl/i) { $cct++; last; }
        }
      }
      if ($k eq 'nodeOS')
      {
	my @oss = split (/ /, $v);
	foreach my $os (@oss)
	{
	  if ($osType eq $os) { $cct++; last; }
	}
      }
      if ($k eq 'nodeCPUType')
      {
        my @cput = split (/ /, $v);
        foreach my $cpu (@cput)
        {
          if ($cpuType eq $cpu) { $cct++; last; }
        }
      }
      $cc++;
    }
    return $q->{ID} if $cc == $cct;
  }
  return 0;
}

1;
