#!/usr/bin/perl -W

use strict;
no strict 'subs';
use warnings;

use IO::Socket::UNIX;
use LWP::UserAgent;
use POSIX qw(strftime);
use File::Copy;
use File::Basename;
use FindBin qw($Bin);
use lib "$Bin";
use XQUtils;
# use Data::Dumper;
my $iniFile;
my $xml;
my %IniSettings=();
my $id;

my $cmd = lc $ARGV[0]
  or die ("Usage: $0 [-i ini_file] add <task_description_file> | kill <task_id> | show\n");

if ($cmd eq '-i') { $iniFile = $ARGV[1]; shift @ARGV; shift @ARGV; $cmd = lc $ARGV[0] }
else { $iniFile = $Bin.'/xqc.ini' }

die ("xqc: Can't read $iniFile!\n") unless -f $iniFile;
XQUtils::parseIniFile ($iniFile, \%IniSettings);
my ($host, $port, $usock) = ($IniSettings{host}, $IniSettings{port}, $IniSettings{socket});
die ("xqc: Bad $iniFile!\n") if (!defined($usock) && !defined($host) && !defined($port));
die ("xqc: Missing host or port in $iniFile!\n") if (!defined($usock) && (!defined($host) || !defined($port)));
die ("xqc: Bad socket $usock!\n") if (defined($usock) && (! -S $usock));

my $user;
$user = (defined($ENV{USER}) ? $ENV{USER} : (defined ($ENV{USERNAME}) ? $ENV{USERNAME} : ''));
die ("xqc: Can't determine user name!\n") if (($user eq '') && (!defined($usock) || ($^O eq 'MSWin32')));

if ($cmd eq 'add')
{
  my %taskX = ();
  my $xf = $ARGV[1] or die ("xqc: Task description file missed in ADD command!\n");
  die ("xqc: Can't read task description file $xf!\n") if (! -r $xf);
  XQUtils::parseIniFile ($xf, \%taskX);
  shift @ARGV; shift @ARGV;
  while (defined($ARGV[0]) && defined($ARGV[1]))
  {
    if ($ARGV[0] ne 'taskArgs')
    {
      $taskX{$ARGV[0]} = $ARGV[1];
      shift @ARGV; shift @ARGV;
    }
    else
    {
      shift @ARGV;
      $taskX{taskArgs} = join (' ', @ARGV);
      last;
    }
  }
  die ("xqc: taskName is missing in parameters!\n") if !exists($taskX{taskName});
  if (exists($taskX{xFile}))
  {
    die ("xqc: Don't know where to copy ('xCopyPath' is missed in $iniFile)!\n") unless defined($IniSettings{copy});
    my $xFile = $IniSettings{copy}.'/'.basename($taskX{xFile}).'.'.time().'.'.$$;
    copy ($taskX{xFile}, $xFile) or die ("xqc: Can't copy $taskX{xFile} to $xFile!\n");
    $taskX{xFile} = $xFile;
  }
  my $td = '';
  while (my ($k,$v) = each(%taskX))
  {
    $td.=' '.$k.'="'.$v.'"';
  }
  $xml = '<add user="'.$user.'"><task'.$td.' /></add>';
}
elsif ($cmd eq 'kill')
{
  die ("xqc: Missing task ID in KILL command!\n") if (!defined($ARGV[1]));
  $id = int ($ARGV[1]);
  $xml = '<kill id="'.$id.'" user="'.$user.'"/>';
}
elsif (($cmd eq 'show') || ($cmd eq 'rshow'))
{
  $xml = '<show user="'.$user.'" />';
}
else
{
  die ("xqc: Unknow command: $cmd!\n");
}

# print STDERR $xml."\n";

my ($code, $tmp, $mess);
if (defined($usock))
{
  my $sock = IO::Socket::UNIX->new(Type => SOCK_STREAM, Peer => $usock) or die "xqc: can't create socket!\n";
  $sock->print ($xml."\n\n");
  while (my $line = $sock->getline)
  {
    if (!defined($code)) { $code = $line } else { $tmp .= $line }
  }
  $sock->close;
  chomp $code;
}
else
{
  my $ua = LWP::UserAgent->new();
  my $resp = $ua->post("http://$host:$port", Content=>$xml) or die ("xqc: Can't connect to $host:$port!");
  ($code, $tmp) = ($resp->code(), $resp->decoded_content());
}

if ($code eq '200')
{
  if ($tmp =~ /<task id=\"(\d+)\" \/>/)
  { $mess = "XQServ: Task was queued with ID=$1." }
  else
  { $mess = "xqc: Unable to parse received data!" }
}
elsif ($code eq '201') { $mess = "XQServ: Task $id was stopped and removed from queue." }
elsif ($code eq '202') { $mess = "XQServ: Task $id was removed from queue." }
elsif ($code eq '203')
{
  $tmp =~ s/<tasks>//;
  $tmp =~ s/<\/tasks>//;
# print STDERR $tmp;
  $mess = ($cmd eq 'show' ? "\nID\tUSER\tTASK\tSTATE\tQTIME\t\t\tSTIME\n".('-' x 75)."\n" : "\nID\tUSER\tTASK\tSERVER\n".('-' x 75)."\n");
  while ($tmp =~ /<task id=\"(\d+)\" name=\"([^\"]+)\" state=\"(.)\" user=\"([^\"]+)\" qtime=\"(\d+)\" stime=\"(\d+)\" sshost=\"([^\"]*)\" ssport=\"(\d+)\" \/>/g)
  {
    if ($cmd eq 'show')
    {
      $mess .="$1\t".substr($4,0,7)."\t".substr($2,0,7)."\t$3\t".strftime("%Y-%m-%d %H:%M:%S",localtime($5))."\t".($6!=0 ? strftime("%Y-%m-%d %H:%M:%S",localtime($6)) : '')."\n";
    }
    elsif (($cmd eq 'rshow') && ($3 eq 'R'))
    {
      $mess .="$1\t".substr($4,0,7)."\t".substr($2,0,7)."\thttp://$7:$8\n";
    }
  }
}
elsif ($code eq '401') { $mess = "XQServ: Task $id was not removed." }
elsif ($code eq '404') { $mess = "XQServ: Can't find task with ID=$id." }
elsif ($code eq '300') { $mess = "XQServ: Not supported." }
elsif ($code eq '500') { $mess = "xqc: Connection refused!" }

print $mess."\n";
